#!/usr/bin/env python
"""
@author: Peter Corke
@author: Jesse Haviland
"""

# 2/8/95  changed D3 to 150.05mm which is closer to data from Lee, AKB86 and
# Tarn fixed errors in COG for links 2 and 3
# 29/1/91 to agree with data from Armstrong etal.  Due to their use
#  of modified D&H params, some of the offsets Ai, Di are
#  offset, and for links 3-5 swap Y and Z axes.
# 14/2/91 to use Paul's value of link twist (alpha) to be consistant
#  with ARCL.  This is the -ve of Lee's values, which means the
#  zero angle position is a righty for Paul, and lefty for Lee.

# all parameters are in SI units: m, radians, kg, kg.m2, N.m, N.m.s etc.

# from math import pi
import numpy as np
from roboticstoolbox import DHRobot, RevoluteDH
from spatialmath import SE3


class Puma560(DHRobot):
    """
    Class that models a Puma 560 manipulator

    :param symbolic: use symbolic constants
    :type symbolic: bool

    ``Puma560()`` is an object which models a Unimation Puma560 robot and
    describes its kinematic and dynamic characteristics using standard DH
    conventions.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.DH.Puma560()
        >>> print(robot)

    Defined joint configurations are:

    - qz, zero joint angle configuration, 'L' shaped configuration
    - qr, vertical 'READY' configuration
    - qs, arm is stretched out in the x-direction
    - qn, arm is at a nominal non-singular configuration

    .. note::
        - SI units are used.
        - The model includes armature inertia and gear ratios.
        - The value of m1 is given as 0 here.  Armstrong found no value for it
          and it does not appear in the equation for tau1 after the
          substituion is made to inertia about link frame rather than COG
          frame.
        - Gravity load torque is the motor torque necessary to keep the joint
          static, and is thus -ve of the gravity caused torque.

    .. warning:: Compared to the MATLAB version of the Toolbox this model
        includes the pedestal, making the z-coordinates 26 inches larger.
        
    :references:
        - "A search for consensus among model parameters reported for the PUMA
          560 robot", P. Corke and B. Armstrong-Helouvry,
          Proc. IEEE Int. Conf. Robotics and Automation, (San Diego),
          pp. 1608-1613, May 1994. (for kinematic and dynamic parameters)
        - "A combined optimization method for solving the inverse kinematics
          problem", Wang & Chen, IEEE Trans. RA 7(4) 1991 pp 489-.
          (for joint angle limits)
        - https://github.com/4rtur1t0/ARTE/blob/master/robots/UNIMATE/puma560/parameters.m

    .. codeauthor:: Peter Corke
    """  # noqa

    def __init__(self, symbolic=False):

        if symbolic:
            import spatialmath.base.symbolic as sym
            zero = sym.zero()
            pi = sym.pi()
        else:
            from math import pi
            zero = 0.0

        deg = pi / 180
        inch = 0.0254

        base = 26.45 * inch    # from mounting surface to shoulder axis

        L = [
            RevoluteDH(
                d=base,       # link length (Dennavit-Hartenberg notation)
                a=0,          # link offset (Dennavit-Hartenberg notation)
                alpha=pi/2,   # link twist (Dennavit-Hartenberg notation)
                I=[0, 0.35, 0, 0, 0, 0],
                # inertia tensor of link with respect to
                # center of mass I = [L_xx, L_yy, L_zz,
                # L_xy, L_yz, L_xz]
                r=[0, 0, 0],
                # distance of ith origin to center of mass [x,y,z]
                # in link reference frame
                m=0,          # mass of link
                Jm=200e-6,    # actuator inertia
                G=-62.6111,   # gear ratio
                B=1.48e-3,    # actuator viscous friction coefficient (measured
                              # at the motor)
                Tc=[0.395, -0.435],
                # actuator Coulomb friction coefficient for
                # direction [-,+] (measured at the motor)
                qlim=[-160*deg, 160*deg]    # minimum and maximum joint angle
            ),

            RevoluteDH(
                d=0, a=0.4318, alpha=zero,
                I=[0.13, 0.524, 0.539, 0, 0, 0],
                r=[-0.3638, 0.006, 0.2275],
                m=17.4,
                Jm=200e-6,
                G=107.815,
                B=.817e-3,
                Tc=[0.126, -0.071],
                qlim=[-110*deg, 110*deg],  # qlim=[-45*deg, 225*deg]
            ),

            RevoluteDH(
                d=0.15005, a=0.0203, alpha=-pi/2,
                I=[0.066, 0.086, 0.0125, 0, 0, 0],
                r=[-0.0203, -0.0141, 0.070],
                m=4.8,
                Jm=200e-6,
                G=-53.7063,
                B=1.38e-3,
                Tc=[0.132, -0.105],
                qlim=[-135*deg, 135*deg]  # qlim=[-225*deg, 45*deg]
            ),

            RevoluteDH(
                d=0.4318, a=0, alpha=pi/2,
                I=[1.8e-3, 1.3e-3, 1.8e-3, 0, 0, 0],
                r=[0, 0.019, 0],
                m=0.82,
                Jm=33e-6,
                G=76.0364,
                B=71.2e-6,
                Tc=[11.2e-3, -16.9e-3],
                qlim=[-266*deg, 266*deg]  # qlim=[-110*deg, 170*deg]
            ),

            RevoluteDH(
                d=0, a=0, alpha=-pi/2,
                I=[0.3e-3, 0.4e-3, 0.3e-3, 0, 0, 0],
                r=[0, 0, 0],
                m=0.34,
                Jm=33e-6,
                G=71.923,
                B=82.6e-6,
                Tc=[9.26e-3, -14.5e-3],
                qlim=[-100*deg, 100*deg]
            ),

            RevoluteDH(
                d=0, a=0, alpha=zero,
                I=[0.15e-3, 0.15e-3, 0.04e-3, 0, 0, 0],
                r=[0, 0, 0.032],
                m=0.09,
                Jm=33e-6,
                G=76.686,
                B=36.7e-6,
                Tc=[3.96e-3, -10.5e-3],
                qlim=[-266*deg, 266*deg]
            )
        ]

        super().__init__(
            L,
            name="Puma 560",
            manufacturer="Unimation",
            keywords=('dynamics', 'symbolic', 'mesh'),
            symbolic=symbolic,
            meshdir="meshes/UNIMATE/puma560"
        )

        # zero angles, L shaped pose
        self.addconfiguration("qz", np.array([0, 0, 0, 0, 0, 0]))

        # ready pose, arm up
        self.addconfiguration("qr", np.array([0, pi/2, -pi/2, 0, 0, 0]))

        # straight and horizontal
        self.addconfiguration("qs", np.array([0, 0, -pi/2, 0, 0, 0]))

        # nominal table top picking pose
        self.addconfiguration("qn", np.array([0, pi/4, pi, 0, pi/4, 0]))

    def ikine_a(self, T, config="lun"):
        """
        Analytic inverse kinematic solution

        :param T: end-effector pose
        :type T: SE3
        :param config: arm configuration, defaults to "lun"
        :type config: str, optional
        :return: joint angle vector in radians
        :rtype: ndarray(6)

        ``robot.ikine_a(T, config)`` is the joint angle vector which achieves the
        end-effector pose ``T```.  The configuration string selects the specific
        solution and is a sting comprising the following letters:

        ======   ==============================================
        Letter   Meaning
        ======   ==============================================
        l        Choose the left-handed configuration
        r        Choose the right-handed configuration
        u        Choose the elbow up configuration
        d        Choose the elbow down configuration
        n        Choose the wrist not-flipped configuration
        f        Choose the wrist flipped configuration
        ======   ==============================================


        :reference:
            - Inverse kinematics for a PUMA 560,
              Paul and Zhang,
              The International Journal of Robotics Research,
              Vol. 5, No. 2, Summer 1986, p. 32-44

        :author: based on MATLAB code by Robert Biro with Gary Von McMurray,
            GTRI/ATRP/IIMB, Georgia Institute of Technology, 2/13/95
 
        """
        def ik3(robot, T, config):

            # solve for the first three joints

            a2 = robot.links[1].a
            a3 = robot.links[2].a
            d1 = robot.links[0].d
            d3 = robot.links[2].d
            d4 = robot.links[3].d

            # The following parameters are extracted from the Homogeneous
            # Transformation as defined in equation 1, p. 34
            
            Px, Py, Pz = T.t
            Pz -= d1  # offset the pedestal height
            theta = np.zeros((3,))

            # Solve for theta[0]
            # r is defined in equation 38, p. 39.
            # theta[0] uses equations 40 and 41, p.39,
            # based on the configuration parameter n1

            r = np.sqrt(Px**2 + Py**2)
            if 'r' in config:
                theta[0] = np.arctan2(Py, Px) + np.arcsin(d3 / r)
            else:
                theta[0] = np.arctan2(Py, Px) + np.pi - np.arcsin(d3 / r)

            # Solve for theta[1]
            # V114 is defined in equation 43, p.39.
            # r is defined in equation 47, p.39.
            # Psi is defined in equation 49, p.40.
            # theta[1] uses equations 50 and 51, p.40, based on the
            # configuration parameter n2
            if 'u' in config:
                n2 = 1
            else:
                n2 = -1

            if 'r' in config:
                n2 = -n2

            V114 = Px * np.cos(theta[0]) + Py * np.sin(theta[0])

            r = np.sqrt(V114**2 + Pz**2)

            Psi = np.arccos(
                (a2**2 - d4**2 - a3**2 + V114**2 + Pz**2)
                / (2.0 * a2 * r))

            if np.isnan(Psi):
                theta = None    # pragma nocover
            else:
                theta[1] = np.arctan2(Pz, V114) + n2 * Psi

                # Solve for theta[2]
                # theta[2] uses equation 57, p. 40.
                num = np.cos(theta[1]) * V114 + np.sin(theta[1]) * Pz - a2
                den = np.cos(theta[1]) * Pz - np.sin(theta[1]) * V114
                theta[2] = np.arctan2(a3, d4) - np.arctan2(num, den)

            return theta


        return self.ikine_6s(T, config, ik3)


if __name__ == '__main__':    # pragma nocover

    puma = Puma560(symbolic=False)
    print(puma)
    print(puma.dyntable())



from roboticstoolbox.models.URDF import *  # noqa
import roboticstoolbox.models.ETS  # noqa
import roboticstoolbox.models.DH  # noqa
from roboticstoolbox.models.list import list


__all__ = [
    'list'
]



"""
@author: Peter Corke
@author: Samuel Drew
"""

from roboticstoolbox import DHRobot, RevoluteDH
from math import pi
import numpy as np


class IRB140(DHRobot):
    """
    Class that models an ABB  IRB140 manipulator

    ``IRB140()`` is a class which models a Unimation Puma560 robot and
    describes its kinematic and dynamic characteristics using standard DH
    conventions.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.DH.IRB140()
        >>> print(robot)

    Defined joint configurations are:

        - qz, zero joint angle configuration
        - qr, vertical 'READY' configuration
        - qd, lower arm horizontal as per data sheet

    .. note:: SI units of metres are used.

    :references:
        - "IRB 140 data sheet", ABB Robotics.
        - "Utilizing the Functional Work Space Evaluation Tool for Assessing a
          System Design and Reconfiguration Alternatives"
          A. Djuric and R. J. Urbanic
        - https://github.com/4rtur1t0/ARTE/blob/master/robots/ABB/IRB140/parameters.m

    .. codeauthor:: Peter Corke
    """  # noqa
    def __init__(self):
        deg = pi/180

        # robot length values (metres)
        d1 = 0.352
        a1 = 0.070
        a2 = 0.360
        d4 = 0.380
        d6 = 0.065

        # Create Links
        # Updated values form ARTE git. Old values left as comments

        L = [
            RevoluteDH(
                d=d1,
                a=a1,
                alpha=-pi/2,
                m=34655.36e-3,
                r=np.array([27.87, 43.12, -89.03])*1e-3,
                I=np.array([512052539.74, 1361335.88, 51305020.72,
                            1361335.88, 464074688.59, 70335556.04,
                            51305020.72, 70335556.04, 462745526.12])*1e-9,
                qlim=[-180 * deg, 180 * deg]
            ),

            RevoluteDH(
                d=0,
                a=a2,
                alpha=0,
                m=15994.59e-3,
                r=np.array([198.29, 9.73, 92.43])*1e03,
                I=np.array([94817914.40, -3859712.77, 37932017.01,
                            -3859712.77, 328604163.24, -1088970.86,
                            37932017.01, -1088970.86, 277463004.88])*1e-9,
                qlim=[-100 * deg, 100 * deg]
            ),

            RevoluteDH(
                d=0,
                a=0,
                alpha=-pi/2,  # alpha=pi/2,
                m=20862.05e-3,
                r=np.array([-4.56, -79.96, -5.86]),
                I=np.array([500060915.95, -1863252.17, 934875.78,
                            -1863252.17, 75152670.69, -15204130.09,
                            934875.78, -15204130.09, 515424754.34])*1e-9,
                qlim=[-220 * deg, 60 * deg]
            ),

            RevoluteDH(
                d=d4,
                a=0,
                alpha=pi/2,  # alpha=-pi/2,
                qlim=[-200 * deg, 200 * deg]
            ),

            RevoluteDH(
                d=0,
                a=0,
                alpha=-pi/2,  # alpha=pi/2,
                qlim=[-120 * deg, 120 * deg]
            ),

            RevoluteDH(
                d=d6,
                a=0,
                alpha=0,  # alpha=pi/2,
                qlim=[-400 * deg, 400 * deg]
            )
        ]

        super().__init__(
            L,
            # basemesh="ABB/IRB140/link0.stl",
            name='IRB 140',
            manufacturer='ABB',
            meshdir="meshes/ABB/IRB140")

        self.addconfiguration("qz", [0, 0, 0, 0, 0, 0])
        self.addconfiguration("qd", [0, -90*deg, 180*deg, 0, 0, -90*deg])
        self.addconfiguration("qr", [0, -90*deg, 90*deg, 0, 90*deg, -90*deg])


if __name__ == '__main__':   # pragma nocover

    robot = IRB140()
    print(robot)



#!/usr/bin/env python

import numpy as np
from roboticstoolbox.robot.ETS import ETS
from roboticstoolbox.robot.ERobot import ERobot
from roboticstoolbox.robot.ELink import ELink


class Frankie(ERobot):
    """
    A class representing the Franka Emika Panda robot arm. ETS taken from [1]
    based on https://frankaemika.github.io/docs/control_parameters.html

    :param et_list: List of elementary transforms which represent the robot
        kinematics
    :type et_list: list of etb.robot.et
    :param q_idx: List of indexes within the ets_list which correspond to
        joints
    :type q_idx: list of int
    :param name: Name of the robot
    :type name: str, optional
    :param manufacturer: Manufacturer of the robot
    :type manufacturer: str, optional
    :param base: Location of the base is the world frame
    :type base: float np.ndarray(4,4), optional
    :param tool: Offset of the flange of the robot to the end-effector
    :type tool: float np.ndarray(4,4), optional
    :param qz: The zero joint angle configuration of the robot
    :type qz: float np.ndarray(7,)
    :param qr: The ready state joint angle configuration of the robot
    :type qr: float np.ndarray(7,)

    References: [1] Kinematic Derivatives using the Elementary Transform
        Sequence, J. Haviland and P. Corke
    """
    def __init__(self):

        deg = np.pi/180
        mm = 1e-3
        tool_offset = (103)*mm

        b0 = ELink(
            v=ETS.rz(),
            name='base0',
            parent=None
        )

        b1 = ELink(
            v=ETS.tx(),
            name='base1',
            parent=b0
        )

        l0 = ELink(
            ETS.tz(0.333),
            ETS.rz(),
            name='link0',
            parent=b1
        )

        l1 = ELink(
            ETS.rx(-90*deg),
            ETS.rz(),
            name='link1',
            parent=l0
        )

        l2 = ELink(
            ETS.rx(90*deg) * ETS.tz(0.316),
            ETS.rz(),
            name='link2',
            parent=l1
        )

        l3 = ELink(
            ETS.tx(0.0825) * ETS.rx(90*deg),
            ETS.rz(),
            name='link3',
            parent=l2
        )

        l4 = ELink(
            ETS.tx(-0.0825) * ETS.rx(-90*deg) * ETS.tz(0.384),
            ETS.rz(),
            name='link4',
            parent=l3
        )

        l5 = ELink(
            ETS.rx(90*deg),
            ETS.rz(),
            name='link5',
            parent=l4
        )

        l6 = ELink(
            ETS.tx(0.088) * ETS.rx(90*deg) * ETS.tz(0.107),
            ETS.rz(),
            name='link6',
            parent=l5
        )

        ee = ELink(
            ETS.tz(tool_offset) * ETS.rz(-np.pi/4),
            name='ee',
            parent=l6
        )

        elinks = [b0, b1, l0, l1, l2, l3, l4, l5, l6, ee]

        super(Frankie, self).__init__(
            elinks,
            name='Frankie',
            manufacturer='Franka Emika, Omron',
            keywords=('mobile',))

        self.addconfiguration(
            "qz", np.array([0, 0, 0, 0, 0, 0, 0, 0, 0]))
        self.addconfiguration(
            "qr", np.array([0, 0, 0, -90, -90, 90, 0, -90, 90]) * deg)


if __name__ == '__main__':   # pragma nocover

    robot = Frankie()
    print(robot)



#!/usr/bin/env python3
"""
@author: Jesse Haviland
@author: Peter Corke
"""
from collections import UserList
from types import SimpleNamespace
import copy
from abc import ABC
import numpy as np
from spatialmath import SE3, SE2
from spatialmath.base import getvector, getunit, trotx, troty, trotz, \
    issymbol, tr2jac, transl2, trot2, removesmall, trinv, verifymatrix, iseye
    
class SuperETS(UserList, ABC):

    # T is a NumPy array (4,4) or None
    # ets_tuple = namedtuple('ETS3', 'eta axis_func axis joint T jindex flip')

    def __init__(
            self, axis=None, eta=None, axis_func=None, 
            unit='rad', j=None, flip=False):

        super().__init__()  # init UserList superclass

        if axis is None and eta is None and axis_func is None:
            # ET()
            # create instance with no values
            self.data = []
            return

        elif isinstance(axis, ETS):
            # copy constructor
            # e = axis_func
            # axis_func = e.axis_func
            # axis = e.axis
            # # et = e.eta
            # j = e.jindex
            # flip = e.isflip
            # joint = e.isjoint
            # T = e.T
            self.data = copy.copy(axis.data)
            return

        if axis in ('Rx', 'Ry', 'Rz', 'tx', 'ty', 'tz'):
            # it's a regular axis

            if eta is None:
                # no value, it's a variable joint
                if unit != 'rad':
                    raise ValueError(
                        'can only use radians for a variable transform')
                joint = True
                T = None

            else:
                # constant value specified
                if not callable(axis_func):
                    raise ValueError('axis func must be callable')
                joint = False
                eta = getunit(eta, unit)
                T = axis_func(eta)
                if j is not None:
                    raise ValueError(
                        'cannot specify joint index for a constant ET')
                if flip:
                    raise ValueError(
                        'cannot specify flip for a constant ET')

        elif axis == 'C':
            # it's a constant element  Ci
            if isinstance(self, ETS):
                # ETS
                if not isinstance(eta, np.ndarray):
                    T = eta.A
                else:
                    T = eta
                if T.shape != (4, 4):
                    raise ValueError('argument must be ndarray(4,4) or SE3')
            else:
                # ETS2
                if not isinstance(eta, np.ndarray):
                    T = eta.A
                else:
                    T = eta
                if T.shape != (3, 3):
                    raise ValueError('argument must be ndarray(3,3) or SE2')
            axis = "C"
            joint = False
            axis_func = None
        else:
            raise ValueError('bad axis specified')

        # Save all the params in a named tuple
        e = SimpleNamespace(eta=eta, axis_func=axis_func, axis=axis, joint=joint, T=T, jindex=j, flip=flip)

        # And make it the only value of this instance
        self.data = [e]

    @property
    def eta(self):
        """
        Get the transform constant

        :return: The constant η if set
        :rtype: float, symbolic or None

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.tx(1)
            >>> e.eta
            >>> e = ETS.rx(90, 'deg')
            >>> e.eta
            >>> e = ETS.ty()
            >>> e.eta

        .. note:: If the value was given in degrees it will be converted and
            stored internally in radians
        """
        return self.data[0].eta

    @eta.setter
    def eta(self, value):
        """
        Set the transform constant

        :param value: The transform constant η
        :type value: float, symbolic or None

        .. note:: No unit conversions are applied, it is assumed to be in
            radians.
        """
        self.data[0].eta = value

    @property
    def axis_func(self):
        return self.data[0].axis_func

    @property
    def axis(self):
        """
        The transform type and axis

        :return: The transform type and axis
        :rtype: str

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.tx(1)
            >>> e.axis
            >>> e = ETS.rx(90, 'deg')
            >>> e.axis

        """
        return self.data[0].axis

    @property
    def n(self):
        """
        Number of joints

        :return: the number of joints in the ETS
        :rtype: int

        Counts the number of joints in the ETS.

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.rx() * ETS.tx(1) * ETS.tz()
            >>> e.n

        :seealso: :func:`joints`
        """
        n = 0
        for et in self:
            if et.isjoint:
                n += 1

        return n

    @property
    def isjoint(self):
        """
        Test if ET is a joint

        :return: True if a joint
        :rtype: bool

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.tx(1)
            >>> e.isjoint
            >>> e = ETS.tx()
            >>> e.isjoint
        """
        return self.data[0].joint

    @property
    def isflip(self):
        """
        Test if ET joint is flipped

        :return: True if joint is flipped
        :rtype: bool

        A flipped joint uses the negative of the joint variable, ie. it rotates
        or moves in the opposite direction.

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.tx()
            >>> e.T(1)
            >>> eflip = ETS.tx(flip=True)
            >>> eflip.T(1)
        """
        return self.data[0].flip

    @property
    def jindex(self):
        """
        Get ET joint index

        :return: The assigmed joint index
        :rtype: int or None

        Allows an ET to be associated with a numbered joint in a robot.

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.tx()
            >>> print(e)
            >>> e = ETS.tx(j=3)
            >>> print(e)
            >>> print(e.jindex)
        """
        return self.data[0].jindex

    @jindex.setter
    def jindex(self, j):
        if not isinstance(j, int) or j < 0:
            raise TypeError(f'jindex is {j}, must be an int >= 0')
        self.data[0].jindex = j

    @property
    def isrevolute(self):
        """
        Test if ET is a rotation

        :return: True if a rotation
        :rtype: bool

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.tx(1)
            >>> e.isrevolute
            >>> e = ETS.rx()
            >>> e.isrevolute
        """
        return self.axis[0] == 'R'

    @property
    def isprismatic(self):
        """
        Test if ET is a translation

        :return: True if a translation
        :rtype: bool

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.tx(1)
            >>> e.isprismatic
            >>> e = ETS.rx()
            >>> e.isprismatic
        """
        return self.axis[0] == 't'

    @property
    def isconstant(self):
        """
        Test if ET is a compilation constant

        :return: True if a compilation constant
        :rtype: bool

        During compilation, consectutive non-joint ETs are compounded/folded
        into a constant transform.  In a ``str`` representation these ETs are
        denoted by ``Ci`` where ``i`` are integers starting at zero.

        .. note:: This is ET is not actually "elementary", it can be a complex
            mix of rotations and translations.

        :seealso: :func:`compile`
        """
        return self.axis[0] == 'C'

    @property
    def config(self):
        """
        Joint configuration string

        :return: A string indicating the joint types
        :rtype: str

        A string comprising the characters 'R' or 'P' which indicate the types
        of joints in order from left to right.

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.tz() * ETS.tx(1) * ETS.rz() * ETS.tx(1)
            >>> e.config

        """
        return ''.join(
            ['R' if self.isrevolute else 'P' for i in self.joints()])

    def joints(self):
        """
        Get index of joint transforms

        :return: indices of transforms that are joints
        :rtype: list of int

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.rz() * ETS.tx(1) * ETS.rz() * ETS.tx(1)
            >>> e.joints()

        """
        return np.where([e.isjoint for e in self])[0]

    def jointset(self):
        """
        Get set of joint indices

        :return: set of unique joint indices
        :rtype: set

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.rz(j=1) * ETS.tx(j=2) * ETS.rz(j=1) * ETS.tx(1)
            >>> e.jointset()
        """
        return set([self[j].jindex for j in self.joints()])

    def T(self, q=None):
        """
        Evaluate an elementary transformation

        :param q: Is used if this ET is variable (a joint)
        :type q: float (radians), required for variable ET's
        :return: The SE(3) or SE(2) matrix value of the ET
        :rtype:  ndarray(4,4) or ndarray(3,3)

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.tx(1)
            >>> e.T()
            >>> e = ETS.tx()
            >>> e.T(2)

        """
        if self.isjoint:
            if self.isflip:
                q = -q
            return self.axis_func(q)
        else:
            return self.data[0].T

    def eval(self, q=None, unit='rad'):
        """
        Evaluate an ETS with joint coordinate substitution

        :param q: joint coordinates
        :type q: array-like
        :param unit: angular unit, "rad" [default] or "deg"
        :type unit: str
        :return: The SE(3) or SE(2) matrix value of the ET sequence
        :rtype:  ndarray(4,4) or ndarray(3,3)

        Effectively the forward-kinematics of the ET sequence.  Compounds the
        transforms left to right, and substitutes in joint coordinates as
        needed from consecutive elements of ``q``.

        .. note:: if ETs have an explicit joint index, this is used to index
            into the vector ``q``.

        .. warning:: do not mix ETs with and without explicit joint index.

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.rz() * ETS.tx(1) * ETS.rz() * ETS.tx(1)
            >>> print(e)
            >>> len(e)
            >>> e[1:3]
            >>> e.eval([0, 0])
            >>> e.eval([90, -90], 'deg')
            >>> e = ETS.tx(j=1) * ETS.ty(j=0)
            >>> e.eval([3, 4])
        """
        if q is not None:
            q = getvector(q, out='list')
        first = True
        for et in self:
            if et.isjoint:
                if et.jindex is None:
                    qj = q.pop(0)
                else:
                    qj = q[et.jindex]
                if et.isrevolute and unit == 'deg':
                    qj *= np.pi / 180.0
                if et.isflip:
                    qj = -qj
                Tk = et.T(qj)
            else:
                # for constants
                Tk = et.T()
            if first:
                T = Tk
                first = False
            else:
                T = T @ Tk

        if isinstance(self, ETS):
            T = SE3(T, check=False)
        elif isinstance(self, ETS2):
            T = SE2(T, check=False)
        T.simplify()
        return T

    def compile(self):
        """
        Compile an ETS

        :return: optimised ETS
        :rtype: ETS

        Perform constant folding for faster evaluation.  Consecutive constant
        ETs are compounded, leading to a constant ET which is denoted by
        ``Ci`` when displayed.

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> robot = rtb.models.ETS.Panda()
            >>> ets = robot.ets()
            >>> ets
            >>> ets.compile()

        :seealso: :func:`isconstant`
        """
        const = None
        ets = ETS()
        for et in self:

            if et.isjoint:
                # a joint
                if const is not None:
                    # flush the constant
                    if not iseye(const):
                        ets *= ETS._CONST(const)
                    const = None
                ets *= et  # emit the joint ET
            else:
                # not a joint
                if const is None:
                    const = et.T()
                else:
                    const = const @ et.T()

        if const is not None:
            # flush the constant, tool transform
            if not iseye(const):
                ets *= ETS._CONST(const)
        return ets

    def __str__(self, q=None):
        """
        Pretty prints the ETS

        :param q: control how joint variables are displayed
        :type q: str
        :return: Pretty printed ETS
        :rtype: str

        ``q`` controls how the joint variables are displayed:

        - None, format depends on number of joint variables
            - one, display joint variable as q
            - more, display joint variables as q0, q1, ...
            - if a joint index was provided, use this value
        - "", display all joint variables as empty parentheses ``()``
        - "θ", display all joint variables as ``(θ)``
        - format string with passed joint variables ``(j, j+1)``, so "θ{0}"
          would display joint variables as θ0, θ1, ... while "θ{1}" would
          display joint variables as θ1, θ2, ...  ``j`` is either the joint
          index, if provided, otherwise a sequential value.

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.rz() * ETS.tx(1) * ETS.rz()
            >>> print(e[:2])
            >>> print(e)
            >>> print(e.__str__(""))
            >>> print(e.__str__("θ{0}"))  # numbering from 0
            >>> print(e.__str__("θ{1}"))  # numbering from 1
            >>> # explicit joint indices
            >>> e = ETS.rz(j=3) * ETS.tx(1) * ETS.rz(j=4)
            >>> print(e)
            >>> print(e.__str__("θ{0}"))

        .. note:: Angular parameters are converted to degrees, except if they
            are symbolic.

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> from spatialmath.base import symbol
            >>> theta, d = symbol('theta, d')
            >>> e = ETS.rx(theta) * ETS.tx(2) * ETS.rx(45, 'deg') * \
            >>>     ETS.ry(0.2) * ETS.ty(d)
            >>> str(e)

        :SymPy: supported
        """
        es = []
        j = 0
        c = 0

        if q is None:
            if len(self.joints()) > 1:
                q = "q{0}"
            else:
                q = "q"

        # For et in the object, display it, data comes from properties
        # which come from the named tuple
        for et in self:

            if et.isjoint:
                if q is not None:
                    if et.jindex is None:
                        _j = j
                    else:
                        _j = et.jindex
                    qvar = q.format(_j, _j+1) # lgtm [py/str-format/surplus-argument]  # noqa
                else:
                    qvar = ""
                if et.isflip:
                    s = f"{et.axis}(-{qvar})"
                else:
                    s = f"{et.axis}({qvar})"
                j += 1

            elif et.isrevolute:
                if issymbol(et.eta):
                    s = f"{et.axis}({et.eta})"
                else:
                    s = f"{et.axis}({et.eta * 180 / np.pi:.4g}°)"

            elif et.isprismatic:
                s = f"{et.axis}({et.eta})"

            elif et.isconstant:
                s = f"C{c}"
                c += 1

            es.append(s)

        return " * ".join(es)

    # redefine * operator to concatenate the internal lists
    def __mul__(self, rest):
        """
        Overloaded ``*`` operator

        :return: [description]
        :rtype: [type]

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e1 = ETS.rz()
            >>> len(e1)
            >>> e2= ETS.tx(2)
            >>> len(e2)
            >>> e = e1 * e2
            >>> len(e)

        .. note:: The ``*`` operator implies composition, but actually the
            result is a new ETS instance that contains the concatenation of
            the left and right operands in an internal list. In this example
            we see the length of the product is 2.
        """
        prod = ETS()
        prod.data = self.data + rest.data
        return prod

    def __imul__(self, rest):
        prod = ETS()
        prod.data = self.data + rest.data
        return prod

    # redefine so that indexing returns an ET type
    def __getitem__(self, i):
        """
        Index or slice an ETS

        :param i: the index or slince
        :type i: int or slice
        :return: Elementary transform (sequence)
        :rtype: ETS

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.rz() * ETS.tx(1) * ETS.rz() * ETS.tx(1)
            >>> e[0]
            >>> e[1]
            >>> e[1:3]

        """
        item = self.__class__()
        data = self.data[i]  # can be [2] or slice, eg. [3:5]
        # ensure that data is always a list
        if isinstance(data, list):
            item.data = data
        else:
            item.data = [data]
        return item

    def pop(self, i=-1):
        """
        Pop value

        :param i: item in the list to pop, default is last
        :type i: int
        :return: the popped value
        :rtype: instance of same type
        :raises IndexError: if there are no values to pop

        Removes a value from the value list and returns it.  The original
        instance is modified.

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.rz() * ETS.tx(1) * ETS.rz() * ETS.tx(1)
            >>> tail = e.pop()
            >>> tail
            >>> e
        """
        item = self.__class__()
        item.data = [super().pop(i)]
        return item

    def insert(self, i=-1, et=None):
        """
        Insert value

        :param i: insert an ET into the ET sequence, default is at the end
        :type i: int
        :param et: the elementary transform to insert
        :type et: ETS

        Inserts an ET into the ET sequence.  The inserted value is at position
        ``i``.

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.rz() * ETS.tx(1) * ETS.rz() * ETS.tx(1)
            >>> f = ETS.ry()
            >>> e.insert(2, f)
            >>> e
        """
        self.data.insert(i, et.data[0])

    def __repr__(self):
        return str(self)

    @classmethod
    def _CONST(cls, T):
        return cls(None, axis='C', eta=T)

    @classmethod
    def SE3(cls, t, rpy=None, tol=100):
        """
        Convert an SE3 to an ETS

        :param t: Translation vector, or an SE(3) matrix
        :type t: array_like(3) or SE3 instance
        :param rpy: roll-pitch-yaw angles in XYZ order
        :type rpy: array_like(3)
        :param tol: Elements small than this many eps are considered as
            being zero, defaults to 100
        :type tol: int, optional
        :return: ET sequence
        :rtype: ETS instance

        Create an ETS from the non-zero translational and rotational
        components.

        - ``SE3(t, rpy)`` convert translation ``t`` and rotation given by XYZ
          roll-pitch-yaw angles ``rpy`` into an ETS.
        - ``SE3(X)`` as above but convert from an SE3 instance ``X``.

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> ETS.SE3(SE3(1,2,3))
            >>> ETS.SE3(SE3.Rx(90, 'deg'))

        .. warning:: ``SE3.rpy()`` is used to determine rotation about the x-,
            y- and z-axes.  For a y-axis rotation with magnitude greater than
            180° this will result in a non-minimal representation with non-zero
            amounts of x- and z-axis rotation.

        :seealso: :func:`~SE3.rpy`
        """
        if isinstance(t, SE3):
            T = t
            t = removesmall(T.t, tol)
            rpy = removesmall(T.rpy(order='zyx'))

        ets = ETS()
        if t[0] != 0:
            ets *= ETS.tx(t[0])
        if t[1] != 0:
            ets *= ETS.ty(t[1])
        if t[2] != 0:
            ets *= ETS.tz(t[2])

        if rpy is not None:
            if rpy[2] != 0:
                ets *= ETS.rz(rpy[2])
            if rpy[1] != 0:
                ets *= ETS.ry(rpy[1])
            if rpy[0] != 0:
                ets *= ETS.rx(rpy[0])

        return ets

    def inv(self):
        r"""
        Inverse of ETS

        :return: [description]
        :rtype: ETS instance

        The inverse of a given ETS.  It is computed as the inverse of the
        individual ETs in the reverse order.

        .. math::

            (\mathbf{E}_0, \mathbf{E}_1 \cdots \mathbf{E}_{n-1} )^{-1} = (\mathbf{E}_{n-1}^{-1}, \mathbf{E}_{n-2}^{-1} \cdots \mathbf{E}_0^{-1}{n-1} )

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.rz(j=2) * ETS.tx(1) * ETS.rx(j=3,flip=True) * ETS.tx(1)
            >>> print(e)
            >>> print(e.inv())
            >>> q = [1,2,3,4]
            >>> print(e.eval(q) * e.inv().eval(q))

        .. note:: It is essential to use explicit joint indices to account for
            the reversed order of the transforms.
        """  # noqa

        inv = ETS()
        for ns in reversed(self.data):
            # get the namespace from the list

            # clone it, and invert the elements to create an inverse
            nsi = copy.copy(ns)
            if nsi.joint:
                nsi.flip ^= True   # toggle flip status
            elif nsi.axis[0] == 'C':
                nsi.T = trinv(nsi.T)
            elif nsi.eta is not None:
                nsi.T = trinv(nsi.T)
                nsi.eta = -nsi.eta
            et = ETS()  # create a new ETS instance
            et.data = [nsi]  # set it's data from the dict
            inv *= et
        return inv


class ETS(SuperETS):
    """
    This class implements an elementary transform sequence (ETS) for 3D

    :param arg: Function to compute ET value
    :type arg: callable
    :param η: The coordinate of the ET. If not supplied the ET corresponds
        to a variable ET which is a joint
    :type η: float, optional
    :param unit: angular unit, "rad" [default] or "deg"
    :type unit: str
    :param j: Explicit joint number within the robot
    :type j: int, optional
    :param flip: Joint moves in opposite direction
    :type flip: bool

    An instance can contain an elementary transform (ET) or an elementary
    transform sequence (ETS). It has list-like properties by subclassing
    UserList, which means we can perform indexing, slicing pop, insert, as well
    as using it as an iterator over its values.

    - ``ETS()`` an empty ETS list
    - ``ETS.XY(η)`` is a constant elementary transform
    - ``ETS.XY(η, 'deg')`` as above but the angle is expressed in degrees
    - ``ETS.XY()`` is a joint variable, the value is left free until evaluation
      time
    - ``ETS.XY(j=J)`` as above but the joint index is explicitly given, this
      might correspond to the joint number of a multi-joint robot.
    - ``ETS.XY(flip=True)`` as above but the joint moves in the opposite sense

    where ``XY`` is one of ``rx``, ``ry``, ``rz``, ``tx``, ``ty``, ``tz``.

    Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.rz(0.3) # a single ET, rotation about z
            >>> len(e)
            >>> e = ETS.rz(0.3) * ETS.tx(2) # an ETS
            >>> len(e)                      # of length 2
            >>> e[1]                        # an ET sliced from the ETS

    :references:
        - Kinematic Derivatives using the Elementary Transform Sequence,
          J. Haviland and P. Corke

    :seealso: :func:`rx`, :func:`ry`, :func:`rz`, :func:`tx`,
        :func:`ty`, :func:`tz`
    """
    @property
    def s(self):
        if self.axis[1] == 'x':
            if self.axis[0] == 'R':
                return np.r_[0, 0, 0, 1, 0, 0]
            else:
                return np.r_[1, 0, 0, 0, 0, 0]
        elif self.axis[1] == 'y':
            if self.axis[0] == 'R':
                return np.r_[0, 0, 0, 0, 1, 0]
            else:
                return np.r_[0, 1, 0, 0, 0, 0]
        else:
            if self.axis[0] == 'R':
                return np.r_[0, 0, 0, 0, 0, 1]
            else:
                return np.r_[0, 0, 1, 0, 0, 0]

    @classmethod
    def rx(cls, eta=None, unit='rad', **kwargs):
        """
        Pure rotation about the x-axis

        :param η: rotation about the x-axis
        :type η: float
        :param unit: angular unit, "rad" [default] or "deg"
        :type unit: str
        :param j: Explicit joint number within the robot
        :type j: int, optional
        :param flip: Joint moves in opposite direction
        :type flip: bool
        :return: An elementary transform
        :rtype: ETS instance

        - ``ETS.rx(η)`` is an elementary rotation about the x-axis by a
          constant angle η
        - ``ETS.rx()`` is an elementary rotation about the x-axis by a variable
          angle, i.e. a revolute robot joint. ``j`` or ``flip`` can be set in
          this case.

        :seealso: :func:`ETS`, :func:`isrevolute`
        :SymPy: supported
        """
        return cls(axis='Rx', eta=eta, axis_func=trotx, unit=unit, **kwargs)

    @classmethod
    def ry(cls, eta=None, unit='rad', **kwargs):
        """
        Pure rotation about the y-axis

        :param η: rotation about the y-axis
        :type η: float
        :param unit: angular unit, "rad" [default] or "deg"
        :type unit: str
        :param j: Explicit joint number within the robot
        :type j: int, optional
        :param flip: Joint moves in opposite direction
        :type flip: bool
        :return: An elementary transform
        :rtype: ETS instance

        - ``ETS.ry(η)`` is an elementary rotation about the y-axis by a
          constant angle η
        - ``ETS.ry()`` is an elementary rotation about the y-axis by a variable
          angle, i.e. a revolute robot joint. ``j`` or ``flip`` can be set in
          this case.

        :seealso: :func:`ETS`, :func:`isrevolute`
        :SymPy: supported
        """
        return cls(axis='Ry', eta=eta, axis_func=troty, unit=unit, **kwargs)

    @classmethod
    def rz(cls, eta=None, unit='rad', **kwargs):
        """
        Pure rotation about the z-axis

        :param η: rotation about the z-axis
        :type η: float
        :param unit: angular unit, "rad" [default] or "deg"
        :type unit: str
        :param j: Explicit joint number within the robot
        :type j: int, optional
        :param flip: Joint moves in opposite direction
        :type flip: bool
        :return: An elementary transform
        :rtype: ETS instance

        - ``ETS.rz(η)`` is an elementary rotation about the z-axis by a
          constant angle η
        - ``ETS.rz()`` is an elementary rotation about the z-axis by a variable
          angle, i.e. a revolute robot joint. ``j`` or ``flip`` can be set in
          this case.

        :seealso: :func:`ETS`, :func:`isrevolute`
        :SymPy: supported
        """
        return cls(axis='Rz', eta=eta, axis_func=trotz, unit=unit, **kwargs)

    @classmethod
    def tx(cls, eta=None, **kwargs):
        """
        Pure translation along the x-axis

        :param η: translation distance along the z-axis
        :type η: float
        :param j: Explicit joint number within the robot
        :type j: int, optional
        :param flip: Joint moves in opposite direction
        :type flip: bool
        :return: An elementary transform
        :rtype: ETS instance

        - ``ETS.tx(η)`` is an elementary translation along the x-axis by a
          distance constant η
        - ``ETS.tx()`` is an elementary translation along the x-axis by a
          variable distance, i.e. a prismatic robot joint. ``j`` or ``flip``
          can be set in this case.

        :seealso: :func:`ETS`, :func:`isprismatic`
        :SymPy: supported
        """

        # this method is 3x faster than using lambda x: transl(x, 0, 0)
        def axis_func(eta):
            return np.array([
                [1, 0, 0, eta],
                [0, 1, 0, 0],
                [0, 0, 1, 0],
                [0, 0, 0, 1]
            ])

        return cls(axis='tx', axis_func=axis_func, eta=eta, **kwargs)

    @classmethod
    def ty(cls, eta=None, **kwargs):
        """
        Pure translation along the y-axis

        :param η: translation distance along the y-axis
        :type η: float
        :param j: Explicit joint number within the robot
        :type j: int, optional
        :param flip: Joint moves in opposite direction
        :type flip: bool
        :return: An elementary transform
        :rtype: ETS instance

        - ``ETS.ty(η)`` is an elementary translation along the y-axis by a
          distance constant η
        - ``ETS.ty()`` is an elementary translation along the y-axis by a
          variable distance, i.e. a prismatic robot joint. ``j`` or ``flip``
          can be set in this case.

        :seealso: :func:`ETS`, :func:`isprismatic`
        :SymPy: supported
        """
        def axis_func(eta):
            return np.array([
                [1, 0, 0, 0],
                [0, 1, 0, eta],
                [0, 0, 1, 0],
                [0, 0, 0, 1]
            ])

        return cls(axis='ty', eta=eta, axis_func=axis_func, **kwargs)

        # return cls(SE3.Ty, axis='ty', eta=eta)

    @classmethod
    def tz(cls, eta=None, **kwargs):
        """
        Pure translation along the z-axis

        :param η: translation distance along the z-axis
        :type η: float
        :param j: Explicit joint number within the robot
        :type j: int, optional
        :param flip: Joint moves in opposite direction
        :type flip: bool
        :return: An elementary transform
        :rtype: ETS instance

        - ``ETS.tz(η)`` is an elementary translation along the z-axis by a
          distance constant η
        - ``ETS.tz()`` is an elementary translation along the z-axis by a
          variable distance, i.e. a prismatic robot joint. ``j`` or ``flip``
          can be set in this case.

        :seealso: :func:`ETS`, :func:`isprismatic`
        :SymPy: supported
        """
        def axis_func(eta):
            return np.array([
                [1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, 1, eta],
                [0, 0, 0, 1]
            ])

        return cls(axis='tz', axis_func=axis_func, eta=eta, **kwargs)

    def jacob0(self, q=None, T=None):
        r"""
        Jacobian in base frame

        :param q: joint coordinates
        :type q: array_like
        :param T: ETS value as an SE(3) matrix if known
        :type T: ndarray(4,4)
        :return: Jacobian matrix
        :rtype: ndarray(6,n)

        ``jacob0(q)`` is the ETS Jacobian matrix which maps joint
        velocity to spatial velocity in the {0} frame.

        End-effector spatial velocity :math:`\nu = (v_x, v_y, v_z, \omega_x,
        \omega_y, \omega_z)^T` is related to joint velocity by
        :math:`{}^{e}\nu = {}^{e}\mathbf{J}_0(q) \dot{q}`.

        If ``ets.eval(q)`` is already computed it can be passed in as ``T`` to
        reduce computation time.

        An ETS represents the relative pose from the {0} frame to the end frame
        {e}. This is the composition of many relative poses, some constant and
        some functions of the joint variables, which we can write as
        :math:`\mathbf{E}(q)`.

        .. math::
        
            {}^0 T_e = \mathbf{E}(q) \in \mbox{SE}(3)

        The temporal derivative of this is the spatial
        velocity :math:`\nu` which is a 6-vector is related to the rate of
        change of joint coordinates by the Jacobian matrix.

        .. math::

           {}^0 \nu = {}^0 \mathbf{J}(q) \dot{q} \in \mathbb{R}^6

        This velocity can be expressed relative to the {0} frame or the {e}
        frame.

        :references:

            - `Kinematic Derivatives using the Elementary Transform Sequence, J. Haviland and P. Corke <https://arxiv.org/abs/2010.08696>`_

        :seealso: :func:`jacobe`, :func:`hessian0`
        """  # noqa

        # TODO what is offset
        # if offset is None:
        #     offset = SE3()

        n = self.n  # number of joints
        q = getvector(q, n)

        if T is None:
            T = self.eval(q)

        # we will work with NumPy arrays for maximum speed
        T = T.A
        U = np.eye(4)  # SE(3) matrix
        j = 0
        J = np.zeros((6, n))

        for et in self:

            if et.isjoint:
                # joint variable
                # U = U @ link.A(q[j], fast=True)
                U = U @ et.T(q[j])

                # TODO???
                # if link == to_link:
                #     U = U @ offset.A

                Tu = trinv(U) @ T
                n = U[:3, 0]
                o = U[:3, 1]
                a = U[:3, 2]
                x = Tu[0, 3]
                y = Tu[1, 3]
                z = Tu[2, 3]

                if et.axis == 'Rz':
                    J[:3, j] = (o * x) - (n * y)
                    J[3:, j] = a

                elif et.axis == 'Ry':
                    J[:3, j] = (n * z) - (a * x)
                    J[3:, j] = o

                elif et.axis == 'Rx':
                    J[:3, j] = (a * y) - (o * z)
                    J[3:, j] = n

                elif et.axis == 'tx':
                    J[:3, j] = n
                    J[3:, j] = np.array([0, 0, 0])

                elif et.axis == 'ty':
                    J[:3, j] = o
                    J[3:, j] = np.array([0, 0, 0])

                elif et.axis == 'tz':
                    J[:3, j] = a
                    J[3:, j] = np.array([0, 0, 0])

                j += 1
            else:
                # constant transform
                U = U @ et.T()

        return J

    def jacobe(self, q=None, T=None):
        r"""
        Jacobian in base frame

        :param q: joint coordinates
        :type q: array_like
        :param T: ETS value as an SE(3) matrix if known
        :type T: ndarray(4,4)
        :return: Jacobian matrix
        :rtype: ndarray(6,n)

        ``jacobe(q)`` is the manipulator Jacobian matrix which maps joint
        velocity to end-effector spatial velocity.

        End-effector spatial velocity :math:`\nu = (v_x, v_y, v_z, \omega_x, \omega_y, \omega_z)^T`
        is related to joint velocity by :math:`{}^{e}\nu = {}^{e}\mathbf{J}_0(q) \dot{q}`.

        If ``ets.eval(q)`` is already computed it can be passed in as ``T`` to
        reduce computation time.

        :seealso: :func:`jacob`, :func:`hessian0`
        """  # noqa

        if T is None:
            T = self.eval(q)

        return tr2jac(T.A) @ self.jacob0(q, T)

    def hessian0(self, q=None, J0=None):
        r"""
        Hessian in base frame

        :param q: joint coordinates
        :type q: array_like
        :param J0: Jacobian in {0} frame
        :type J0: ndarray(6,n)
        :return: Hessian matrix
        :rtype: ndarray(6,n,n)

        This method calculcates the Hessisan of the ETS. One of ``J0`` or
        ``q`` is required. If ``J0`` is already calculated for the joint
        coordinates ``q`` it can be passed in to to save computation time

        An ETS represents the relative pose from the {0} frame to the end frame
        {e}. This is the composition of many relative poses, some constant and
        some functions of the joint variables, which we can write as
        :math:`\mathbf{E}(q)`.

        .. math::

            {}^0 T_e = \mathbf{E}(q) \in \mbox{SE}(3)

        The temporal derivative of this is the spatial
        velocity :math:`\nu` which is a 6-vector is related to the rate of
        change of joint coordinates by the Jacobian matrix.

        .. math::

            {}^0 \nu = {}^0 \mathbf{J}(q) \dot{q} \in \mathbb{R}^6

        This velocity can be expressed relative to the {0} frame or the {e}
        frame.

        The temporal derivative of spatial velocity is spatial acceleration,
        which again can be expressed with respect to the {0} or {e} frames

        .. math::

            {}^0 \dot{\nu} = \mathbf{J}(q) \ddot{q} + \dot{\mathbf{J}}(q) \dot{q} \in \mathbb{R}^6 \\
                      &= \mathbf{J}(q) \ddot{q} + \dot{q}^T \mathbf{H}(q) \dot{q}

        The manipulator Hessian tensor :math:`H` maps joint velocity to
        end-effector spatial acceleration, expressed in the {0} coordinate
        frame.

        :references:
            - `Kinematic Derivatives using the Elementary Transform Sequence, J. Haviland and P. Corke <https://arxiv.org/abs/2010.08696>`_

        :seealso: :func:`jacob0`
        """  # noqa

        n = self.n

        if J0 is None:
            if q is None:
                q = np.copy(self.q)
            else:
                q = getvector(q, n)

            J0 = self.jacob0(q)
        else:
            verifymatrix(J0, (6, n))

        H = np.zeros((6, n, n))

        for j in range(n):
            for i in range(j, n):

                H[:3, i, j] = np.cross(J0[3:, j], J0[:3, i])
                H[3:, i, j] = np.cross(J0[3:, j], J0[3:, i])

                if i != j:
                    H[:3, j, i] = H[:3, i, j]

        return H


class ETS2(SuperETS):
    """
    This class implements an elementary transform sequence (ETS) for 2D

    :param arg: Function to compute ET value
    :type arg: callable
    :param η: The coordinate of the ET. If not supplied the ET corresponds
        to a variable ET which is a joint
    :type η: float, optional
    :param unit: angular unit, "rad" [default] or "deg"
    :type unit: str
    :param j: Explicit joint number within the robot
    :type j: int, optional
    :param flip: Joint moves in opposite direction
    :type flip: bool

    An instance can contain an elementary transform (ET) or an elementary
    transform sequence (ETS). It has list-like properties by subclassing
    UserList, which means we can perform indexing, slicing pop, insert, as well
    as using it as an iterator over its values.

    - ``ETS()`` an empty ETS list
    - ``ETS.XY(η)`` is a constant elementary transform
    - ``ETS.XY(η, 'deg')`` as above but the angle is expressed in degrees
    - ``ETS.XY()`` is a joint variable, the value is left free until evaluation
      time
    - ``ETS.XY(j=J)`` as above but the joint index is explicitly given, this
      might correspond to the joint number of a multi-joint robot.
    - ``ETS.XY(flip=True)`` as above but the joint moves in the opposite sense

    where ``XY`` is one of ``r``, ``tx``, ``ty``.

    Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS2 as ETS
            >>> e = ETS.r(0.3)  # a single ET, rotation about z
            >>> len(e)
            >>> e = ETS.r(0.3) * ETS.tx(2)  # an ETS
            >>> len(e)                      # of length 2
            >>> e[1]                        # an ET sliced from the ETS

    :references:
        - Kinematic Derivatives using the Elementary Transform Sequence,
          J. Haviland and P. Corke

    :seealso: :func:`r`, :func:`tx`, :func:`ty`
    """

    @classmethod
    def r(cls, eta=None, unit='rad', **kwargs):
        """
        Pure rotation

        :param η: rotation angle
        :type η: float
        :param unit: angular unit, "rad" [default] or "deg"
        :type unit: str
        :param j: Explicit joint number within the robot
        :type j: int, optional
        :param flip: Joint moves in opposite direction
        :type flip: bool
        :return: An elementary transform
        :rtype: ETS instance

        - ``ETS.r(η)`` is an elementary rotation by a constant angle η
        - ``ETS.r()`` is an elementary rotation by a variable angle, i.e. a
          revolute robot joint. ``j`` or ``flip`` can be set in
          this case.

        .. note:: In the 2D case this is rotation around the normal to the
            xy-plane.

        :seealso: :func:`ETS`, :func:`isrevolute`
        """
        return cls(axis='R', eta=eta, 
            axis_func=lambda theta: trot2(theta), unit=unit, **kwargs)

    @classmethod
    def tx(cls, eta=None, **kwargs):
        """
        Pure translation along the x-axis

        :param η: translation distance along the z-axis
        :type η: float
        :param j: Explicit joint number within the robot
        :type j: int, optional
        :param flip: Joint moves in opposite direction
        :type flip: bool
        :return: An elementary transform
        :rtype: ETS instance

        - ``ETS.tx(η)`` is an elementary translation along the x-axis by a
          distance constant η
        - ``ETS.tx()`` is an elementary translation along the x-axis by a
          variable distance, i.e. a prismatic robot joint. ``j`` or ``flip``
          can be set in this case.

        :seealso: :func:`ETS`, :func:`isprismatic`
        """
        return cls(axis='tx', eta=eta, 
            axis_func=lambda x: transl2(x, 0), **kwargs)

    @classmethod
    def ty(cls, eta=None, **kwargs):
        """
        Pure translation along the y-axis

        :param η: translation distance along the y-axis
        :type η: float
        :param j: Explicit joint number within the robot
        :type j: int, optional
        :param flip: Joint moves in opposite direction
        :type flip: bool
        :return: An elementary transform
        :rtype: ETS instance

        - ``ETS.tx(η)`` is an elementary translation along the y-axis by a
          distance constant η
        - ``ETS.tx()`` is an elementary translation along the y-axis by a
          variable distance, i.e. a prismatic robot joint. ``j`` or ``flip``
          can be set in this case.

        :seealso: :func:`ETS`
        """
        return cls(axis='ty', eta=eta, 
            axis_func=lambda y: transl2(0, y), **kwargs)


if __name__ == "__main__":

    # print(ETS.rx(0.2))
    # print(ETS.rx(45, 'deg'))
    # print(ETS.tz(0.75))
    # e = ETS.rx(45, 'deg') * ETS.tz(0.75)
    # print(e)
    # print(e.eval())

    # from roboticstoolbox import ETS
    # e = ETS.rz() * ETS.tx(1) * ETS.rz() * ETS.tx(1)
    # print(e.eval([0, 0]))
    # print(e.eval([90, -90], 'deg'))
    # a = e.pop()
    # print(a)

    # from spatialmath.base import symbol

    # theta, d = symbol('theta, d')

    # e = ETS.rx(theta) * ETS.tx(2) * ETS.rx(45, 'deg') * ETS.ry(0.2) * ETS.ty(d)
    # print(e)

    # e = ETS()
    # e *= ETS.rx()
    # e *= ETS.tz()
    # print(e)

    # print(e.__str__("θ{0}"))
    # print(e.__str__("θ{1}"))

    # e = ETS.rx() * ETS._CONST(SE3()) * ETS.tx(0.3)
    # print(e)

    # l1 = 0.672
    # l2 = -0.2337
    # l3 = 0.4318
    # l4 = 0.0203
    # l5 = 0.0837
    # l6 = 0.4318

    # e = ETS.tz(l1) * ETS.rz() * ETS.ry() * ETS.ty(l2) * ETS.tz(l3) * ETS.ry() \
    #     * ETS.tx(l4) * ETS.ty(l5) * ETS.tz(l6) * ETS.rz() * ETS.ry() * ETS.rz()
    # print(e.joints())
    # print(e.config)
    # print(e.eval(np.zeros(6)))
    # ec = e.compile()
    # print(ec)
    # print(ec.eval(np.zeros(6)))

    # print(ETS.SE3(SE3.Rz(200, 'deg')))

    # a = ETS.rx()
    # b = ETS(a)
    # print(b)
    # a = ETS.tz()
    # print(b)
    # e = ETS.rz(j=5) * ETS.tx(1) * ETS.rx(j=7, flip=True) * ETS.tx(1)
    # print(e)

    # print(e.inv())

    # q = [1, 2, 3, 4, 5, 6, 7, 8]
    # print(e.eval(q))
    # print(e.inv().eval(q))
    # print(e.eval(q) * e.inv().eval(q))


    e = ETS.rz() * ETS.tx(-1) * ETS.tx(1) * ETS.rz()
    print(e)
    print(e.compile())


import numpy as np
from spatialmath import SE3
import roboticstoolbox as rtb
import timeit
from ansitable import ANSITable, Column

# change for the robot IK under test, must set:
#  * robot, the DHRobot object
#  * T, the end-effector pose
#  * q0, the initial joint angles for solution

example = 'puma'  # 'panda'

if example == 'puma':
    # Puma robot case
    robot = rtb.models.DH.Puma560()
    q = robot.qn
    q0 = robot.qz
    T = robot.fkine(q)
elif example == 'panda':
    # Panda robot case
    robot = rtb.models.DH.Panda()
    T = SE3(0.7, 0.2, 0.1) * SE3.OA([0, 1, 0], [0, 0, -1])
    q0 = robot.qz


# build the list of IK methods to test
# it's a tuple:
#  - the method to execute
#  - the name for the results table
#  - the statement to execute for timeit
ikfuncs = [ 
    (robot.ikine_LM,  # Levenberg-Marquadt
        "ikine_LM",
        "sol = robot.ikine_LM(T, q0)"
    ),
    (robot.ikine_LMS, # Levenberg-Marquadt (Sugihara)
        "ikine_LMS",
        "sol = robot.ikine_LMS(T, q0)"
    ),
    (robot.ikine_min, #numerical solution with no constraints
        "ikine_min(qlim=False)",
        "sol = robot.ikine_min(T, q0)"
    ),
    (lambda T, q0: robot.ikine_min(T, q0, qlim=True), #numerical solution with constraints
        "ikine_min(qlim=True)",
        "sol = robot.ikine_min(T, q0, qlim=True)"
    ),
    # (robot.ikine_mmc, #numerical solution with no constraints
    #     "ikine_min(qlim=False)",
    #     "sol = robot.ikine_min(T, q0)"
    # ),
]
if hasattr(robot, "ikine_a"):
    a =  (robot.ikine_a, # analytic solution
        "ikine_a",
        "sol = robot.ikine_a(T)"
    )
    ikfuncs.insert(0, a)    

# setup to run timeit
setup = '''
from __main__ import robot, T, q0
'''
N = 10

# setup results table
table = ANSITable(
    Column("Operation", headalign="^", colalign='<'),
    Column("Time (ms)", headalign="^", fmt="{:.2g}"),
    Column("Error", headalign="^", fmt="{:.3g}"),
    border="thick")

# test the IK methods
for ik in ikfuncs:
    print('Testing:', ik[1])
    
    # test the method, don't pass q0 to the analytic function
    if ik[1] == "ikine_a":
        sol = ik[0](T)
    else:
        sol = ik[0](T, q0=q0)

    # print error message if there is one
    if not sol.success:
        print('  failed:', sol.reason)

    # evalute the error
    err = np.linalg.norm(T - robot.fkine(sol.q))
    print('  error', err)

    if N > 0:
        # evaluate the execution time
        t = timeit.timeit(stmt=ik[2], setup=setup, number=N)
    else:
        t = 0

    # add it to the output table
    table.row(f"`{ik[1]}`", t/N*1e3, err)

# pretty print the results     
table.print()
print(table.markdown())


"""
@author: Gautam Sinha, Indian Institute of Technology, Kanpur
  (original MATLAB version)
@author: Peter Corke
@author: Samuel Drew
"""

from roboticstoolbox import DHRobot, RevoluteDH
from math import pi


class KR5(DHRobot):
    """
    Class that models a Kuka KR5 manipulator

    ``KR5()`` is a class which models a Kuka KR5 robot and
    describes its kinematic characteristics using standard DH
    conventions.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.DH.KR5()
        >>> print(robot)

    Defined joint configurations are:

      - qk1, nominal working position 1
      - qk2, nominal working position 2
      - qk3, nominal working position 3

    .. note::
      - SI units of metres are used.
      - Includes an 11.5cm tool in the z-direction

    :references:
      - https://github.com/4rtur1t0/ARTE/blob/master/robots/KUKA/KR5_arc/parameters.m

    .. codeauthor:: Gautam Sinha, Indian Institute of Technology, Kanpur (original MATLAB version)
    .. codeauthor:: Samuel Drew
    .. codeauthor:: Peter Corke
    """  # noqa

    def __init__(self):
        deg = pi / 180

        # Updated values form ARTE git. Old values left as comments

        L1 = RevoluteDH(a=0.18, d=0.4,
                        alpha=-pi/2,  # alpha=pi / 2,
                        qlim=[-155 * deg, 155 * deg]
                        )
        L2 = RevoluteDH(a=0.6, d=0,  # d=0.135,
                        alpha=0,  # alpha=pi,
                        qlim=[-180 * deg, 65 * deg]
                        )
        L3 = RevoluteDH(a=0.12,
                        d=0,  # d=0.135,
                        alpha=pi/2,  # alpha=-pi / 2,
                        qlim=[-15 * deg, 158 * deg]
                        )
        L4 = RevoluteDH(a=0.0,
                        d=-0.62,  # d=0.62,
                        alpha=-pi/2,  # alpha=pi / 2,
                        qlim=[-350 * deg, 350 * deg]
                        )
        L5 = RevoluteDH(a=0.0,
                        d=0.0,
                        alpha=pi/2,  # alpha=-pi / 2,
                        qlim=[-130 * deg, 130 * deg]
                        )
        L6 = RevoluteDH(a=0,
                        d=-0.115,
                        alpha=pi,
                        qlim=[-350 * deg, 350 * deg]
                        )

        L = [L1, L2, L3, L4, L5, L6]

        # Create SerialLink object
        super().__init__(
            L,
            # meshdir="KUKA/KR5_arc",
            name='KR5',
            manufacturer='KUKA',
            meshdir="meshes/KUKA/KR5_arc")

        self.addconfiguration("qz", [0, 0, 0, 0, 0, 0])
        self.addconfiguration(
          "qk1", [pi / 4, pi / 3, pi / 4, pi / 6, pi / 4, pi / 6])
        self.addconfiguration(
          "qk2", [pi / 4, pi / 3, pi / 6, pi / 3, pi / 4, pi / 6])
        self.addconfiguration(
          "qk3", [pi / 6, pi / 3, pi / 6, pi / 3, pi / 6, pi / 3])


if __name__ == '__main__':   # pragma nocover
    robot = KR5()
    print(robot)



"""
@author Samuel Drew
"""


class Element:     # pragma nocover

    TX = 0
    TY = 1
    TZ = 2
    RX = 3
    RY = 4
    RZ = 5
    DH_STANDARD = 6
    DH_MODIFIED = 7

    # an array of counters for the application of each rule
    # just for debugging.
    rules = [0] * 20

    # mapping from type to string
    typeName = ["TX", "TY", "TZ", "RX", "RY", "RZ", "DH", "DHm"]

    # order of elementary transform for each DH convention
    # in each tuple, the first element is the transform type,
    # the second is true if it can be a joint variable.
    dhStandard = (RZ, 1), (TX, 0), (TZ, 1), (RX, 0)
    dhModified = (RX, 0), (TX, 0), (RZ, 1), (TZ, 1)

    def __init__(
            self,
            elementIn=None,
            stringIn=None,
            eltype=None,
            constant=None,
            sign=0):

        self.var = None         # eg. q1, for joint var types
        self.symconst = None    # eg. L1, for lengths

        # DH parameters, only set if type is DH_STANDARD/MODIFIED
        self.theta = None
        self.alpha = None
        self.A = None
        self.D = None
        self.prismatic = None
        self.offset = None

        if stringIn:
            if elementIn or eltype or constant:
                raise ValueError(
                    "if parsing a string, string must be the only input")
            i = None
            sType = stringIn[0:2]     # Tx, Rx etc
            sRest = stringIn[2:]       # the argument including brackets

            if not (sRest[-1] == ")" and sRest[0] == "("):
                raise ValueError("brackets")

            match = False
            for i in range(6):
                check = self.typeName[i].lower()
                if sType.lower() == check:
                    match = True
                    self.eltype = i
            if not match:
                raise ValueError("bad transform name: " + sType)

            sRest = sRest[1:-1]     # get the argument from between brackets

            # handle an optional minus sign
            negative = ""

            if sRest[0] == '-':
                negative = "-"
                sRest = sRest[1]

            if sRest[0] == "q":
                self.var = negative + sRest
            elif sRest[0] == 'L':
                self.symconst = negative + sRest
            else:
                try:
                    constant = float(sRest)
                    if negative == "-":
                        constant = -constant
                        # negative = ""
                except Exception:
                    raise ValueError("bad argument in term " + stringIn)

        elif elementIn:
            if not isinstance(elementIn, Element):
                raise TypeError("elementIn must be an existing Element object")
            self.eltype = elementIn.eltype
            if elementIn.var:
                self.var = elementIn.var
            if elementIn.symconst:
                self.symconst = elementIn.symconst
            self.constant = elementIn.constant
            # if sign < 0:
            #     self.negate()

        # one of TX, TY ... RZ, DH_STANDARD/MODIFIED
        if eltype:
            self.eltype = eltype
            if constant:
                self.constant = constant    # eg. 90, for angles
            if sign < 0:
                self.negate()

    @staticmethod
    def showRuleUsage():
        for i in range(20):
            if Element.rules[i] > 0:
                print("Rule " + str(i) + ": " + str(Element.rules[i]))

    def istrans(self):
        return (self.eltype == self.TX) or (self.eltype == self.TY) \
            or (self.eltype == self.TZ)

    def isrot(self):
        return (self.eltype == self.RX) or (self.eltype == self.RY) or \
            (self.eltype == self.RZ)

    def isjoint(self):
        return self.var is not None

    def axis(self):
        if self.eltype == self.RX or self.eltype == self.TX:
            return 0
        elif self.eltype == self.RY or self.eltype == self.TY:
            return 1
        elif self.eltype == self.RZ or self.eltype == self.TZ:
            return 2
        else:
            raise ValueError("bad transform type")

    def symAdd(self, s1, s2):

        if s1 is None and s2 is None:
            return None
        elif s1 and s2 is None:
            return s1
        elif s1 is None and s2:
            return s2
        else:
            if s2[0] == "-":
                return s1 + s2
            else:
                return s1 + "+" + s2

    def add(self, e):
        if self.eltype != Element.DH_STANDARD and \
                self.eltype != Element.DH_MODIFIED:
            raise ValueError("wrong element type " + str(self))
        print("  adding: " + str(self) + " += " + str(e))
        if e.eltype == self.RZ:
            if e.isjoint():
                self.prismatic = 0
                self.var = e.var
                self.offset = e.constant
                self.theta = 0.0
            else:
                self.theta = e.constant
        elif e.eltype == self.TX:
            self.A = e.symconst
        elif e.eltype == self.TZ:
            if e.isjoint():
                self.prismatic = 1
                self.var = e.var
                self.D = None
            else:
                self.D = e.symconst
        elif e.eltype == self.RX:
            self.alpha = e.constant
        else:
            raise ValueError("Can't factorise " + str(e))

    # test if this particular element could be part of a DH term
    # eg. Rz(q1) can be, Rx(q1) cannot.
    def factorMatch(self, dhWhich, i, verbose):

        dhFactors = None
        match = False

        if dhWhich == self.DH_STANDARD:
            dhFactors = self.dhStandard
        elif dhWhich == self.DH_MODIFIED:
            dhFactors = self.dhModified
        else:
            raise ValueError("bad DH type")

        match = (self.eltype == dhFactors[i][0]) and not (
            (dhFactors[i][1] == 0) and self.isjoint())

        if verbose > 0:
            print(" matching " + str(self) + " (i=" + str(i) + ") " +
                  " to " + self.typeName[dhFactors[i][0]] + "<" +
                  str(dhFactors[i][1]) + ">" + " -> " + str(match))
        return match

    def merge(self, e):
        assert type(e) == Element, "merge(Element e)"
        """
        don't merge if dissimilar transform or
        both are joint variables
        """
        if e.eltype != self.eltype or e.isjoint() and self.isjoint():
            return self

        sum = Element(self)

        sum.var = self.symAdd(self.var, e.var)
        sum.symconst = self.symAdd(self.symconst, e.symconst)
        sum.constant = self.constant + e.constant

        if not sum.isjoint() and sum.symconst is None and sum.constant == 0:
            print("Eliminate: " + self + " " + e)
            return None
        else:
            print("Merge: " + self + " " + e + " := " + sum)
            return sum

    def swap(self, next, dhWhich):
        assert type(next) == Element, "type(next) == Element"

        # don't swap if both are joint variables
        if self.isjoint() and next.isjoint():
            return False

        if dhWhich == Element.DH_STANDARD:
            # order = [2, 0, 3, 4, 0, 1]
            if self.eltype == Element.TZ and next.eltype == Element.TX or \
                    self.eltype == Element.TX and next.eltype == Element.RX \
                    and next.isjoint() or \
                    self.eltype == Element.TY and next.eltype == Element.RY \
                    and next.isjoint() or \
                    self.eltype == Element.TZ and next.eltype == Element.RZ \
                    and next.isjoint() or \
                    not self.isjoint() and self.eltype == Element.RX and \
                    next.eltype == Element.TX or \
                    not self.isjoint() and self.eltype == Element.RY and \
                    next.eltype == Element.TY or \
                    not self.isjoint() and not next.isjoint() and \
                    self.eltype == Element.TZ and \
                    next.eltype == Element.RZ or \
                    self.eltype == Element.TY and \
                    next.eltype == Element.TZ or \
                    self.eltype == Element.TY and next.eltype == Element.TX:
                print("Swap: " + self + " <-> " + next)
                return True
        elif dhWhich == Element.DH_MODIFIED:
            if self.eltype == Element.RX and next.eltype == Element.TX or \
                    self.eltype == Element.RY and \
                    next.eltype == Element.TY or \
                    self.eltype == Element.RZ and \
                    next.eltype == Element.TZ or \
                    self.eltype == Element.TZ and next.eltype == Element.TX:
                print("Swap: " + self + " <-> " + next)
                return True
        else:
            raise ValueError("bad DH type")
        return False

    # negate the arguments of the element
    def negate(self):

        self.constant = -self.constant

        if self.symconst:
            s = list(self.symconst)
            # if no leading sign character insert one (so we can flip it)
            if s[0] != "+" and s[0] != "-":
                s.insert(0, "+")
            for i in range(len(s)):
                if s[i] == "+":
                    s[i] = "-"
                elif s[i] == "-":
                    s[i] = "+"
                if s[0] == "+":
                    s.pop(0)
        s = "".join(s)    # lgtm [py/unused-local-variable]

    '''
    Return a string representation of the parameters (argument)
    of the element, which can be a number, symbolic constant,
    or a joint variable.
    '''
    def argString(self):
        s = ""

        if self.eltype == Element.RX or Element.RY or Element.RZ or \
                Element.TX or Element.TY or Element.TZ:
            if self.var:
                s = self.var
            if self.symconst:
                if self.var:
                    if self.symconst[0] != "-":
                        s = s + "+"
                s = s + self.symconst
            # constants always displayed with a sign character
            if self.constant != 0.0:
                if self.constant >= 0.0:
                    s = s + "+" + '{0:.3f}'.format(self.constant)
                else:
                    s = s + '{0:.3f}'.format(self.constant)
        elif self.eltype == Element.DH_STANDARD or Element.DH_MODIFIED:
            # theta, d, a, alpha
            # theta
            if self.prismatic == 0:
                # revolute joint
                s = s + self.var
                if self.offset >= 0:
                    s = s + "+" + '{0:.3f}'.format(self.offset)
                elif self.offset < 0:
                    s = s + '{0:.3f}'.format(self.offset)
            else:
                # prismatic joint
                s = s + '{0:.3f}'.format(self.theta)
            s = s + ", "

            # d
            if self.prismatic > 0:
                s = s + self.var
            else:
                s = s + self.D if self.D else s + "0"
            s = s + ", "

            # a
            s = s + self.A if self.A else s + "0"
            s = s + ", "

            # alpha
            s = s + '{0:.3f}'.format(self.alpha)
        else:
            raise ValueError("bad Element type")
        return s

    def toString(self):
        s = Element.typeName[self.eltype] + "("
        s = s + self.argString()
        s = s + ")"
        return s



#!/usr/bin/env python
"""
@author Jesse Haviland
"""

from roboticstoolbox.backends.Connector import Connector
import roboticstoolbox as rp
import numpy as np
import spatialmath as sm
import time
from queue import Queue

_sw = None
sw = None


def _import_swift():     # pragma nocover
    import importlib
    global sw
    try:
        sw = importlib.import_module('swift')
        # from swift import start_servers
    except ImportError:
        print(
            '\nYou must install the python package swift, see '
            'https://github.com/jhavl/swift\n')
        raise


class Swift(Connector):  # pragma nocover
    """
    Graphical backend using Swift

    Swift is a web app built on three.js. It supports many 3D graphical
    primitives including meshes, boxes, ellipsoids and lines. It can render
    Collada objects in full color.

    :param realtime: Force the simulator to display no faster than real time,
        note that it may still run slower due to complexity
    :type realtime: bool
    :param display: Do not launch the graphical front-end of the simulator.
        Will still simulate the robot. Runs faster due to not needing to
        display anything.
    :type display: bool

    Example:

    .. code-block:: python
        :linenos:

        import roboticstoolbox as rtb

        robot = rtb.models.DH.Panda()  # create a robot

        pyplot = rtb.backends.Swift()   # create a Swift backend
        pyplot.add(robot)              # add the robot to the backend
        robot.q = robot.qz             # set the robot configuration
        pyplot.step()                  # update the backend and graphical view

    :references:

        - https://github.com/jhavl/swift

    """
    def __init__(self, realtime=True, display=True):
        super(Swift, self).__init__()

        self.sim_time = 0.0
        self.robots = []
        self.shapes = []
        self.outq = Queue()
        self.inq = Queue()

        self.realtime = realtime
        self.display = display

        self.recording = False

        if self.display and sw is None:
            _import_swift()

    #
    #  Basic methods to do with the state of the external program
    #

    def launch(self, browser=None):
        """
        Launch a graphical backend in Swift by default in the default browser
        or in the specified browser

        :param browser: browser to open in: one of
            'google-chrome', 'chrome', 'firefox', 'safari', 'opera'
            or see for full list
            https://docs.python.org/3.8/library/webbrowser.html#webbrowser.open_new
        :type browser: string

        ``env = launch(args)`` create a 3D scene in a running Swift instance as
        defined by args, and returns a reference to the backend.

        """

        super().launch()

        if self.display:
            sw.start_servers(self.outq, self.inq, browser=browser)
            self.last_time = time.time()

    def step(self, dt=0.05):
        """
        Update the graphical scene

        :param dt: time step in seconds, defaults to 0.05
        :type dt: int, optional

        ``env.step(args)`` triggers an update of the 3D scene in the Swift
        window referenced by ``env``.

        .. note::

            - Each robot in the scene is updated based on
              their control type (position, velocity, acceleration, or torque).
            - Upon acting, the other three of the four control types will be
              updated in the internal state of the robot object.
            - The control type is defined by the robot object, and not all
              robot objects support all control types.
            - Execution is blocked for the specified interval

        """

        # TODO how is the pose of shapes updated prior to step?

        super().step

        self._step_robots(dt)
        self._step_shapes(dt)

        # Adjust sim time
        self.sim_time += dt

        # Send updated sim time to Swift
        if self.display:

            # If realtime is set, delay progress if we are running too quickly
            if self.realtime:
                time_taken = (time.time() - self.last_time)
                diff = dt - time_taken

                if diff > 0:
                    time.sleep(diff)

                self.last_time = time.time()

            self._draw_all()
            self._send_socket('sim_time', self.sim_time)

    def reset(self):
        """
        Reset the graphical scene

        ``env.reset()`` triggers a reset of the 3D scene in the Swift window
        referenced by ``env``. It is restored to the original state defined by
        ``launch()``.

        """

        super().reset

    def restart(self):
        """
        Restart the graphics display

        ``env.restart()`` triggers a restart of the Swift view referenced by
        ``env``. It is closed and relaunched to the original state defined by
        ``launch()``.

        """

        super().restart

    def close(self):
        """
        Close the graphics display

        ``env.close()`` gracefully disconnectes from the Swift visualizer
        referenced by ``env``.
        """

        super().close()

    #
    #  Methods to interface with the robots created in other environemnts
    #

    def add(
            self, ob, show_robot=True, show_collision=False,
            readonly=False):
        """
        Add a robot to the graphical scene

        :param ob: the object to add
        :type ob: Robot or Shape
        :param show_robot: Show the robot visual geometry,
            defaults to True
        :type show_robot: bool, optional
        :param show_collision: Show the collision geometry,
            defaults to False
        :type show_collision: bool, optional
        :return: object id within visualizer
        :rtype: int
        :param readonly: If true, swif twill not modify any robot attributes,
            the robot is only nbeing displayed, not simulated,
            defaults to False
        :type readonly: bool, optional

        ``id = env.add(robot)`` adds the ``robot`` to the graphical
            environment.

        .. note::

            - Adds the robot object to a list of robots which will be updated
              when the ``step()`` method is called.

        """
        # id = add(robot) adds the robot to the external environment. robot
        # must be of an appropriate class. This adds a robot object to a
        # list of robots which will act upon the step() method being called.

        # TODO can add more than a robot right?

        super().add()

        if isinstance(ob, rp.ERobot):
            robot = ob.to_dict()
            robot['show_robot'] = show_robot
            robot['show_collision'] = show_collision

            robot_object = {
                'ob': ob,
                'readonly': readonly
            }

            if self.display:
                id = self._send_socket('robot', robot)

                loaded = 0
                while loaded == 0:
                    loaded = int(self._send_socket('is_loaded', id))
                    time.sleep(0.1)
            else:
                id = len(self.robots)

            self.robots.append(robot_object)
            return id
        elif isinstance(ob, rp.Shape):
            shape = ob.to_dict()
            if self.display:
                id = self._send_socket('shape', shape)
            else:
                id = len(self.shapes)
            self.shapes.append(ob)
            return id

    def remove(self):
        """
        Remove a robot to the graphical scene

        ``env.remove(robot)`` removes the ``robot`` from the graphical
            environment.
        """

        # TODO - shouldn't this have an id argument? which robot does it remove
        # TODO - it can remove any entity?

        super().remove()

    def hold(self):           # pragma: no cover
        '''
        hold() keeps the browser tab open i.e. stops the browser tab from
        closing once the main script has finished.

        '''

        while True:
            time.sleep(1)

    def start_recording(self, file_name, framerate, format='webm'):
        """
        Start recording the canvas in the Swift simulator

        :param file_name: The file name for which the video will be saved as
        :type file_name: string
        :param framerate: The framerate of the video - to be timed correctly,
            this should equalt 1 / dt where dt is the time supplied to the
            step function
        :type framerate: float
        :param format: This is the format of the video, one of 'webm', 'gif',
            'png', or 'jpg'
        :type format: string

        ``env.start_recording(file_name)`` starts recording the simulation
            scene and will save it as file_name once
            ``env.start_recording(file_name)`` is called
        """

        valid_formats = ['webm', 'gif', 'png', 'jpg']

        if format not in valid_formats:
            raise ValueError(
                "Format can one of 'webm', 'gif', 'png', or 'jpg'")

        if not self.recording:
            self._send_socket(
                'start_recording', [framerate, file_name, format])
            self.recording = True
        else:
            raise ValueError(
                "You are already recording, you can only record one video"
                " at a time")

    def stop_recording(self):
        """
        Start recording the canvas in the Swift simulator. This is optional
        as the video will be automatically saved when the python script exits

        ``env.stop_recording()`` stops the recording of the simulation, can
            only be called after ``env.start_recording(file_name)``
        """

        if self.recording:
            self._send_socket('stop_recording')
        else:
            raise ValueError(
                "You must call swift.start_recording(file_name) before trying"
                " to stop the recording")

    def _step_robots(self, dt):

        for robot_object in self.robots:
            robot = robot_object['ob']

            if robot_object['readonly'] or robot.control_type == 'p':
                pass            # pragma: no cover

            elif robot.control_type == 'v':

                for i in range(robot.n):
                    robot.q[i] += robot.qd[i] * (dt)

                    if np.any(robot.qlim[:, i] != 0) and \
                            not np.any(np.isnan(robot.qlim[:, i])):
                        robot.q[i] = np.min([robot.q[i], robot.qlim[1, i]])
                        robot.q[i] = np.max([robot.q[i], robot.qlim[0, i]])

            elif robot.control_type == 'a':
                pass

            else:            # pragma: no cover
                # Should be impossible to reach
                raise ValueError(
                    'Invalid robot.control_type. '
                    'Must be one of \'p\', \'v\', or \'a\'')

    def _step_shapes(self, dt):

        for shape in self.shapes:

            T = shape.base
            t = T.t
            r = T.rpy('rad')

            t += shape.v[:3] * (dt)
            r += shape.v[3:] * (dt)

            shape.base = sm.SE3(t) * sm.SE3.RPY(r)

    def _draw_all(self):

        for i in range(len(self.robots)):
            self._send_socket(
                'robot_poses', [i, self.robots[i]['ob'].fk_dict()])

        for i in range(len(self.shapes)):
            self._send_socket(
                'shape_poses', [i, self.shapes[i].fk_dict()])

    def _send_socket(self, code, data=None):
        msg = [code, data]

        self.outq.put(msg)
        return self.inq.get()



#PoseGraph Pose graph 

import roboticstoolbox as rtb
import pgraph
from spatialmath import base, SE2
import matplotlib.pyplot as plt
import numpy as np
import scipy as sp
import zipfile
import time
import math

class PoseGraph:
    
    # properties
    #     graph
        
    #     ngrid
    #     center
    #     cellsize

    def __init__(self, filename, laser=False, verbose=False):
        # parse the file data
        # we assume g2o format
        #    VERTEX* vertex_id X Y THETA
        #    EDGE* startvertex_id endvertex_id X Y THETA IXX IXY IYY IXT IYT ITT
        # vertex numbers start at 0
        
        self.laser = laser
        
        self.graph = pgraph.UGraph(verbose=verbose)

        path = rtb.path_to_datafile(filename)

        if filename.endswith('.zip'):
            zf = zipfile.ZipFile(path, 'r')
            opener = zf.open
            filename = filename[:-4]
        else:
            opener = open
            filename = path

        with opener(filename, 'r') as f:

            toroformat = False
            nlaser = 0
            
            # indices into ROBOTLASER1 record for the 3x3 info matrix in column major
            # order
            g2o =  [0,  1,  2,  1,  3,  4,   2,  4,  5]
            toro = [0,  1, 4,  1,  2,  5,  4,  5,   3]
            
            # we keep an array self. = vindex(gi) to map g2o vertex index to PGraph vertex index
            
            vindex = {}
            firstlaser = True
            for line in f:
                # for zip file, we get data as bytes not str
                if isinstance(line, bytes):
                    line = line.decode()
                
                # is it a comment?
                if line.startswith('#'):
                    continue
                
                tokens = line.split(' ')
                
                # g2o format records
                if tokens[0] == 'VERTEX_SE2':
                    v = self.graph.add_vertex([float(x) for x in tokens[2:5]])
                    id = int(tokens[1])
                    vindex[id] = v
                    v.id = id

                    v.type = 'vertex'
                        
                elif tokens[0] ==  'VERTEX_XY':
                    v = self.graph.add_vertex([float(x) for x in tokens[2:4]])
                    id = int(tokens[1])
                    vindex[id] = v
                    v.id = id
                    v.type = 'landmark'
                        
                elif tokens[0] ==  'EDGE_SE2':                    
                    v1 = vindex[int(tokens[1])]
                    v2 = vindex[int(tokens[2])]

                    # create the edge
                    e = self.graph.add_edge(v1, v2)

                    # create the edge data as a structure
                    #  X  Y  T
                    #  3  4  5
                    e.mean = np.array([float(x) for x in tokens[3:6]])
                    
                    # IXX IXY IXT IYY IYT ITT
                    #   6   7   8   9  10  11
                    info = np.array([float(x) for x in tokens[6:12]])
                    e.info = np.reshape(info[g2o], (3,3))
                        
                ## TORO format records
                elif tokens[0] == 'VERTEX2':
                    toroformat = True
                    v = self.graph.add_vertex([float(x) for x in tokens[2:5]])
                    id = int(tokens[1])
                    vindex[id] = v
                    v.id = id
                    v.type = 'vertex'
                        
                elif tokens[0] == 'EDGE2':
                    toroformat = True
                    v1 = vindex[int(tokens[1])]
                    v2 = vindex[int(tokens[2])]

                    # create the edge
                    e = self.graph.add_edge(v1, v2)

                    # create the edge data as a structure
                    #  X  Y  T
                    #  3  4  5
                    e.mean = [float(x) for x in tokens[3:6]]
                    
                    # IXX IXY IXT IYY IYT ITT
                    #   6   7   8   9  10  11
                    info = np.array([float(x) for x in tokens[6:12]])
                    e.info = np.reshape(info[toro], (3,3))
                        
                elif tokens[0] == 'ROBOTLASER1':
                    if not laser:
                        continue
                    
                    # laser records are associated with the immediately preceding VERTEX record

                    # not quite sure what all the fields are
                    # 1 ?
                    # 2 min scan angle
                    # 3 scan range
                    # 4 angular increment
                    # 5 maximum range possible
                    # 6 ?
                    # 7 ?
                    # 8 N = number of beams
                    # 9 to 9+N laser range data
                    # 9+N+1 ?
                    # 9+N+2 ?
                    # 9+N+3 ?
                    # 9+N+4 ?
                    # 9+N+5 ?
                    # 9+N+6 ?
                    # 9+N+7 ?
                    # 9+N+8 ?
                    # 9+N+9 ?
                    # 9+N+10 ?
                    # 9+N+11 ?
                    # 9+N+12 timestamp (*nix timestamp)
                    # 9+N+13 laser type (str)
                    # 9+N+14  ?

                    if firstlaser:
                        nbeams = int(tokens[8])
                        lasermeta = tokens[2:6]
                        firstlaser = False

                    v.theta = np.arange(0, nbeams) * float(tokens[4]) + float(tokens[2])
                    v.range = np.array([float(x) for x in tokens[9:nbeams+9]])
                    v.time = float(tokens[21+nbeams])
                    nlaser+= 1
                else:
                    raise RuntimeError(f"Unexpected line  {line} in {filename}")

            if toroformat:
                print(f"loaded TORO/LAGO format file: {self.graph.n} nodes, {self.graph.ne} edges")
            else:
                print(f"loaded g2o format file: {self.graph.n} nodes, {self.graph.ne} edges")
            if nlaser > 0:
                lasermeta = [float(x) for x in lasermeta]
                
                self._angmin = lasermeta[0]
                self._angmax = sum(lasermeta[0:2])
                self._maxrange = lasermeta[3]

                fov = np.degrees([self._angmin, self._angmax])
                print(f"  {nlaser} laser scans: {nbeams} beams, fov {fov[0]:.1f}° to {fov[1]:.1f}°, max range {self._maxrange}")

        self.vindex = vindex
    
    def scan(self, i):
        v = self.vindex[i]
        return v.range, v.theta
    
    def scanxy(self, i):
        v = self.vindex[i]
        
        range, theta = self.scan(i)
        x = range * np.cos(theta)
        y = range * np.sin(theta)

        return np.c_[x,y]
    
    def plot_scan(self, n):
        n = base.getvector(n)
        for i in n:
            x, y = self.scanxy(i)
            plt.plot(x, y, '.', 'MarkerSize', 10)
            plt.pause(1)

    def pose(self, i):
        return self.vindex[i].coord
    
    def time(self, i):
        return self.vindex[i].time
    
    def plot(self, **kwargs):
        self.graph.plot(**kwargs)
        plt.xlabel('x')
        plt.ylabel('y')
        plt.grid(True)
    
    def scanmap(self, centre=(75,50), ngrid=3000, cellsize=0.1, maxrange=None):
        
        self._centre = centre
        self._cellsize = cellsize
        self._ngrid = ngrid

        # h = waitbar(0, 'rendering a map')
        
        world = np.zeros((ngrid, ngrid), np.int32)
        for i in range(0, self.graph.n):
            
            # if i % 20 == 0:
            #     waitbar(i/self.graph.n, h)
            
            xy = self.scanxy(i)
            r,theta = self.scan(i)
            if maxrange is not None:
                xy = np.delete(xy, r > maxrange, axis=0)
            xyt = self.vindex[i].coord
            
            xy = SE2(xyt) * xy.T
            
            # start of each ray
            p1 = self.w2g(xyt[0:2])
            
            for end in xy.T:
                
                # end of each ray
                p2 = self.w2g(end[0:2])
                
                # all cells along the ray
                x, y = bresenham(p1, p2)

                try:
                    # decrement cells along the ray, these are free space
                    world[y[:-1],x[:-1]] = world[y[:-1],x[:-1]] - 1
                    # increment target cell
                    world[y[-1],x[-1]] = world[y[-1],x[-1]] + 1
                except IndexError:
                    # silently ignore rays to points outside the grid map
                    pass

        return world
    
    def w2g(self, w):        
        return np.round((w - self._centre) / self._cellsize) + self._ngrid / 2

    def g2w(self, g):
        return ( np.r_[g] - self._ngrid / 2) * self._cellsize + self._centre

    def plot_occgrid(self, w, block=True):
        
        bl = self.g2w([0,0])
        tr = self.g2w([w.shape[1], w.shape[0]])

        w = np.where(w < 0, -1, w)  # free cells are -1
        w = np.where(w > 0, 1, w)   # occupied cells are +1
        w = -w 

        # plot it
        plt.imshow(w, cmap='gray', extent=[bl[0], tr[0], bl[1], tr[1]])
        plt.xlabel('x')
        plt.ylabel('y')
        plt.show(block=block)

def bresenham(p1, p2):
    
    # ensure all values are integer
    x1 = round(p1[0]); y1 = round(p1[1])
    x2 = round(p2[0]); y2 = round(p2[1])

    # compute the vertical and horizontal change
    dx = abs(x2 - x1)
    dy = abs(y2 - y1)
    steep = dy > dx
    
    if steep:
        # if slope > 1 swap the deltas
        t = dx
        dx = dy
        dy = t

    # The main algorithm goes here.
    if dy == 0:
        q = np.zeros((dx+1,1), np.int64)
    else:
        q = np.arange(np.floor(dx / 2), -dy * (dx + 1) + np.floor(dx / 2), -dy)
        q = np.r_[0, np.diff(np.mod(q, dx)) >= 0]
            
    if steep:
        if y1 <= y2:
            y = np.arange(y1, y2 + 1)
        else:
            y = np.arange(y1, y2 - 1, -1)

        if x1 <= x2:
            x = x1 + np.cumsum(q)
        else:
            x = x1 - np.cumsum(q)
    else:
        if x1 <= x2:
            x = np.arange(x1, x2 + 1)
        else:
            x = np.arange(x1, x2 - 1, -1)

        if y1 <= y2:
            y = y1 + np.cumsum(q)
        else:
            y = y1 - np.cumsum(q)
    
    return x, y




    
    
#   This source code is part of the graph optimization package
#   deveoped for the lectures of robotics2 at the University of Freiburg.
#
#     Copyright (c) 2007 Giorgio Grisetti, Gian Diego Tipaldi
#
#   It is licences under the Common Creative License,
#   Attribution-NonCommercial-ShareAlike 3.0
#
#   You are free:
#     - to Share - to copy, distribute and transmit the work
#     - to Remix - to adapt the work
#
#   Under the following conditions:
#
#     - Attribution. You must attribute the work in the manner specified
#       by the author or licensor (but not in any way that suggests that
#       they endorse you or your use of the work).
#
#     - Noncommercial. You may not use this work for commercial purposes.
#
#     - Share Alike. If you alter, transform, or build upon this work,
#       you may distribute the resulting work only under the same or
#       similar license to this one.
#
#   Any of the above conditions can be waived if you get permission
#   from the copyright holder.  Nothing in this license impairs or
#   restricts the author's moral rights.
#
#   This software is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied
#   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#   PURPOSE.
    
    
    # %ls-slam.m
    # %this file is released under the creative common license
    
    # solves a graph-based slam problem via least squares
    # vmeans: matrix containing the column vectors of the poses of the vertices
    # 	 the vertices are odrered such that vmeans[i] corresponds to the ith id
    # eids:	 matrix containing the column vectors [idFrom, idTo]' of the ids of the vertices
    # 	 eids[k] corresponds to emeans[k] and einfs[k].
    # emeans: matrix containing the column vectors of the poses of the edges
    # einfs:  3d matrix containing the information matrices of the edges
    # 	 einfs(:,:,k) refers to the information matrix of the k-th edge.
    # n:	 number of iterations
    # newmeans: matrix containing the column vectors of the updated vertices positions
    
    def optimize(self, iterations = 10, animate = False, retain = False):
        
        
        g2  =  PGraph(self.graph)  # deep copy
        
        eprev  =  math.inf
        for i in range(iterations):
            if animate:
                if not retain:
                    plt.clf()
                g2.plot()
                plt.pause(0.5)
            
            vmeans, energy = linearize_and_solve(g2)
            g2.setcoord(vmeans)
            
            if energy >= eprev:
                break
            eprev = energy
        
        self.graph = g2

        return g2


#computes the taylor expansion of the error function of the k_th edge
#vmeans: vertices positions
#eids:   edge ids
#emeans: edge means
#k:	 edge number
#e:	 e_k(x)
#A:	 d e_k(x) / d(x_i)
#B:	 d e_k(x) / d(x_j)
#function [e, A, B] = linear_factors(vmeans, eids, emeans, k)
def linear_factors(self, edge):
    # extract the ids of the vertices connected by the kth edge
    # 	id_i = eids(1,k)
    # 	id_j = eids(2,k)
    # extract the poses of the vertices and the mean of the edge
    #     v_i = vmeans(:,id_i)
    #     v_j = vmeans(:,id_j)
    #     z_ij = emeans(:,k)

    v  =  g.vertices(edge)
    v_i  =  g.coord(v(1))
    v_j  =  g.coord(v(2))
    z_ij  =  g.edata(edge).mean

    # compute the homoeneous transforms of the previous solutions
    zt_ij = base.trot2(z_ij[2], t=z_ij[0:2])
    vt_i = base.trot2(v_i[2], t=v_i[0:2])
    vt_j = base.trot2(v_j[2], t=v_j[0:2])

    # compute the displacement between x_i and x_j

    f_ij = base.trinv2(vt_i @ vt_j)

    # this below is too long to explain, to understand it derive it by hand
    theta_i = v_i[3]
    ti = v_i[0:2,0]
    tj = v_j[0:2,0]
    dt_ij = tj - ti

    si = sin(theta_i)
    ci = cos(theta_i)

    A =  np.array([
        [-ci, -si, [-si,  ci] @ dt_ij],
        [ si, -ci, [-ci, -si] @ dt_ij],
        [  0,   0, -1 ],
            ])
    B  = np.array([
        [  ci, si, 0],
        [ -si, ci, 0],
        [  0,   0, 1 ],
            ])

    ztinv = base.trinv2(zt_ij)
    T = ztinv @ f_ij
    e = np.r_[base.transl2(T), base.angle(T)]
    ztinv[0:2,2] =  0
    A = ztinv @ A
    B = ztinv @ B

    return e, A, B 



# linearizes and solves one time the ls-slam problem specified by the input
# vmeans:   vertices positions at the linearization point
# eids:     edge ids
# emeans:   edge means
# einfs:    edge information matrices
# newmeans: new solution computed from the initial guess in vmeans
def linearize_and_solve(self):
    t0  =  time.time()
    print('solving')
    
    # H and b are respectively the system matrix and the system vector
    H = np.zeros((self.n * 3, self.n * 3))
    b = zeros(g.n * 3,1)
    # this loop constructs the global system by accumulating in H and b the contributions
    # of all edges (see lecture)
    fprintf('.')
    
    etotal  =  0
    for edge  in self.edges:
        
        e, A, B = self.linear_factors(edge)
        omega  =  g.edata(edge).info
        # compute the blocks of H^k
        
        # not quite sure whey SE3 is being transposed, what does that mean?
        b_i   =  -A.T @ omega @ e
        b_j   =  -B.T @ omega @ e
        H_ii  =   A.T @ omega @ A
        H_ij  =   A.T @ omega @ B
        H_jj  =   B.T @ omega @ B
        
        v = g.vertices(edge)
        
        i = v[0]; j = v[1]

        islice = slice(i*3, (i+1)*3)
        jslice = slice(j*3, (j+1)*3)

        # accumulate the blocks in H and b

        H[islice, islice] += Hii
        H[jslice, jslice] += Hjj
        H[islice, jslice] += Hij
        H[jslice, islice] += Hij.T

        b[islice,0] += b_i
        b[jslice,0] += b_j

        etotal  =  etotal + np.inner(e, e)
        printf('.', end=None)

    
    # %note that the system (H b) is obtained only from
    # %relative constraints. H is not full rank.
    # %we solve the problem by anchoring the position of
    # %the the first vertex.
    # %this can be expressed by adding the equation
    #  deltax(1:3,1) = 0
    # which is equivalent to the following

    H[0:3,0:3] += np.eye(3)
    
    SH = sp.bsr_sparse(H)
    print('.', end=None)
    deltax = sp.spsolve(SH, b)  # SH \ b
    print('.', end=None)
    
    # split the increments in nice 3x1 vectors and sum them up to the original matrix
    newmeans  =  self.coord() + np.reshape(deltax, (3, self.n))
    
    # normalize the angles between -PI and PI
    # for (i = 1:size(newmeans,2))
    #     s = sin(newmeans(3,i))
    #     c = cos(newmeans(3,i))
    #     newmeans(3,i) = atan2(s,c)
    newmeans[3,:] = base.angdiff(newmeans[3,:])

    dt  =  time.time() - t0
    print(f"done in {dt:0.2f} sec.  Total cost {etotal}")


    return newmeans, energy

#   This source code is part of the graph optimization package
#   deveoped for the lectures of robotics2 at the University of Freiburg.
#
#     Copyright (c) 2007 Giorgio Grisetti, Gian Diego Tipaldi
#
#   It is licences under the Common Creative License,
#   Attribution-NonCommercial-ShareAlike 3.0
#
#   You are free:
#     - to Share - to copy, distribute and transmit the work
#     - to Remix - to adapt the work
#
#   Under the following conditions:
#
#     - Attribution. You must attribute the work in the manner specified
#       by the author or licensor (but not in any way that suggests that
#       they endorse you or your use of the work).
#
#     - Noncommercial. You may not use this work for commercial purposes.
#
#     - Share Alike. If you alter, transform, or build upon this work,
#       you may distribute the resulting work only under the same or
#       similar license to this one.
#
#   Any of the above conditions can be waived if you get permission
#   from the copyright holder.  Nothing in this license impairs or
#   restricts the author's moral rights.
#
#   This software is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied
#   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#   PURPOSE.



# #computes the pose vector v from an homogeneous transform A
# function v = t2v(A)
#     v(1:2, 1) = A(1:2,3)
#     v(3,1) = atan2(A(2,1),A(1,1))
# end


if __name__ == "__main__":

    scan = PoseGraph('killian.g2o', laser=True, verbose=False)
    print(scan.graph.nc)
    # scan.plot()
    w = scan.scanmap(maxrange=40)
    scan.plot_occgrid(w)


import numpy as np
import math
from collections import namedtuple
import matplotlib.pyplot as plt
from spatialmath.base.argcheck import isvector, \
    getvector, assertmatrix


def tpoly(q0, qf, t, qd0=0, qdf=0):
    """
    Generate scalar polynomial trajectory

    :param q0: initial value
    :type q0: float
    :param qf: final value
    :type qf: float
    :param t: time
    :type t: float or array_like
    :param qd0: initial velocity, optional
    :type q0: float
    :param qdf: final velocity, optional
    :type q0: float
    :return: trajectory
    :rtype: namedtuple

    - ``tg = tpoly(q0, q1, t)`` is a scalar trajectory (Mx1) that varies
      smoothly from ``q0`` to ``qf`` using a quintic polynomial.  The initial
      and final velocity and acceleration are zero. Time ``t`` can be either:

        * an integer scalar, indicating the total number of timesteps

            - Velocity is in units of distance per trajectory step, not per
              second.
            - Acceleration is in units of distance per trajectory step squared,
              *not* per second squared.

        * an array_like, containing the time steps.

            - Results are scaled to units of time.

    - ``tg = tpoly(q0, q1, t, qd0, qdf)`` as above but specify the initial and
      final velocity. The initial and final acceleration are zero.

    The return value is a namedtuple (named ``tpoly``) with elements:

        - ``x``  the time coordinate as a numpy ndarray, shape=(M,)
        - ``y``  the position as a numpy ndarray, shape=(M,)
        - ``yd``  the velocity as a numpy ndarray, shape=(M,)
        - ``ydd``  the acceleration as a numpy ndarray, shape=(M,)

    .. note:: The time vector T is assumed to be monotonically increasing, and
        time scaling is based on the first and last element.

    References:

    - Robotics, Vision & Control, Chap 3,
      P. Corke, Springer 2011.

    :seealso: :func:`lspb`, :func:`t1plot`, :func:`jtraj`.
    """

    if isinstance(t, int):
        t = np.arange(0, t)
        istime = False
    elif isvector(t):
        t = getvector(t)
        istime = True
    else:
        raise TypeError('bad argument for time, must be int or vector')

    tf = max(t)
    # solve for the polynomial coefficients using least squares
    X = [
            [0,             0,           0,           0,       0,   1],
            [tf ** 5,       tf ** 4,     tf ** 3,     tf ** 2, tf,  1],
            [0,             0,           0,           0,       1,   0],
            [5 * tf ** 4,   4 * tf ** 3, 3 * tf ** 2, 2 * tf,  1,   0],
            [0,             0,           0,           2,       0,   0],
            [20 * tf ** 3, 12 * tf ** 2, 6 * tf,      2,       0,   0]
    ]
    coeffs, resid, rank, s = np.linalg.lstsq(
        X, np.r_[q0, qf, qd0, qdf, 0, 0], rcond=None)

    # coefficients of derivatives
    coeffs_d = coeffs[0:5] * np.arange(5, 0, -1)
    coeffs_dd = coeffs_d[0:4] * np.arange(4, 0, -1)

    # evaluate the polynomials
    p = np.polyval(coeffs, t)
    pd = np.polyval(coeffs_d, t)
    pdd = np.polyval(coeffs_dd, t)

    return namedtuple('tpoly', 'x y yd ydd istime')(t, p, pd, pdd, istime)


# -------------------------------------------------------------------------- #

def lspb(q0, qf, t, V=None):
    """
    Scalar trapezoidal trajectory

    :param q0: initial value
    :type q0: float
    :param qf: final value
    :type qf: float
    :param t: time
    :type t: float or array_like
    :param V: velocity of linear segment, optional
    :type V: float
    :return: trajectory
    :rtype: namedtuple

    Computes a trapezoidal trajectory, which has a linear motion segment with
    parabolic blends.

    - ``tg = lspb(q0, qf, t)`` is a scalar trajectory (Mx1) that varies
      smoothly from ``q0`` to ``qf`` in M steps using a constant velocity
      segment and parabolic blends.  Time ``t`` can be either:

        * an integer scalar, indicating the total number of timesteps

            - Velocity is in units of distance per trajectory step, not per
              second.
            - Acceleration is in units of distance per trajectory step squared,
              *not* per second squared.

        * an array_like, containing the time steps.

            - Results are scaled to units of time.

    - ``tg = lspb(q0, q1, t, V)``  as above but specifies the velocity of the
      linear segment which is normally computed automatically.

    The return value is a namedtuple (named ``lspb``) with elements:

        - ``x``  the time coordinate as a numpy ndarray, shape=(M,)
        - ``y``  the position as a numpy ndarray, shape=(M,)
        - ``yd``  the velocity as a numpy ndarray, shape=(M,)
        - ``ydd``  the acceleration as a numpy ndarray, shape=(M,)

    .. note::

        - For some values of V no solution is possible and an error is flagged.
        - The time vector, if given, is assumed to be monotonically increasing,
          and time scaling is based on the first and last element.

    :References:

        - Robotics, Vision & Control, Chap 3,
        P. Corke, Springer 2011.

    :seealso: :func:`tpoly`, :func:`t1plot`, :func:`jtraj`.
    """

    if isinstance(t, int):
        t = np.arange(0, t)
        istime = False
    elif isvector(t):
        t = getvector(t)
        istime = True
    else:
        raise TypeError('bad argument for time, must be int or vector')

    tf = max(t)

    if V is None:
        # if velocity not specified, compute it
        V = (qf - q0) / tf * 1.5
    else:
        V = abs(V) * np.sign(qf - q0)
        if abs(V) < (abs(qf - q0) / tf):
            raise ValueError('V too small')
        elif abs(V) > (2 * abs(qf-q0) / tf):
            raise ValueError('V too big')

    if q0 == qf:
        # Commented these because they arent used anywhere
        # s = np.ones((len(t), len(t))) @ q0
        # sd = np.zeros((len(t), len(t)))
        # sdd = np.zeros((len(t), len(t)))
        return

    tb = (q0 - qf + V * tf) / V
    a = V / tb

    p = np.zeros((len(t),))
    pd = np.zeros((len(t),))
    pdd = np.zeros((len(t),))

    for i, _t in enumerate(t):

        if _t <= tb:
            # initial blend
            p[i] = q0 + a/2 * _t ** 2
            pd[i] = a * _t
            pdd[i] = a
        elif _t <= (tf - tb):
            # linear motion
            p[i] = (qf + q0 - V * tf) / 2 + V * _t
            pd[i] = V
            pdd[i] = 0
        else:
            # final blend
            p[i] = qf - a / 2 * tf ** 2 + a * tf * _t - a / 2 * _t ** 2
            pd[i] = a * tf - a * _t
            pdd[i] = -a

    return namedtuple(
        'lspb', 'x y yd ydd xblend istime')(t, p, pd, pdd, tb, istime)


# -------------------------------------------------------------------------- #

def jtraj(q0, qf, tv, qd0=None, qd1=None):
    """
    Compute a joint-space trajectory

    :param q0: initial joint coordinate
    :type q0: N-element array_like
    :param qf: final joint coordinate
    :type qf: N-element array_like
    :param tv: time vector or number of steps
    :type tv: array_like or int
    :param qd0: initial velocity, defaults to zero
    :type qd0: N-element array_like, optional
    :param qd1: final velocity, defaults to zero
    :type qd1: N-element array_like, optional
    :return: trajectory of coordinates and optionally velocity and acceleration
    :rtype: namedtuple

    - ``tg = jtraj(q0, qf, M)`` is a joint space trajectory where the joint
      coordinates vary from ``q0`` (N) to ``qf`` (N).  A quintic (5th order)
      polynomial is used with default zero boundary conditions for velocity and
      acceleration.
      Time is assumed to vary from 0 to 1 in ``M`` steps.

    - ``tg = jtraj(q0, qf, M, qd0, qdf)`` as above but also specifies initial
      ``qd0`` (N) and final ``qdf`` (N) joint velocity for the trajectory.

    - ``tg = jtraj(q0, qf, tv)``, as above but the number of steps in the
      trajectory is defined by the length of the time vector ``tv`` (M).

    - ``tg = jtraj(q0, qf, tv, qd0, qdf)`` as above but specifies initial and
      final joint velocity for the trajectory and a time vector.

    The return value is a namedtuple (named ``jtraj``) with elements:

        - ``t``  the time coordinate as a numpy ndarray, shape=(M,)
        - ``q``  the position as a numpy ndarray, shape=(M,N)
        - ``qd``  the velocity as a numpy ndarray, shape=(M,N)
        - ``qdd``  the acceleration as a numpy ndarray, shape=(M,N)

    Notes:

    - When a time vector is provided the velocity and acceleration outputs
      are scaled assuming that the time vector starts at zero and increases
      linearly.

    :seealso: :func:`ctraj`, :func:`qplot`, :func:`~SerialLink.jtraj`
    """
    # print(f"  --- jtraj: {q0} --> {q1} in {tv}")
    if isinstance(tv, int):
        tscal = 1.0
        t = np.linspace(0, 1, tv)  # normalized time from 0 -> 1
    else:
        tscal = max(tv)
        t = tv.flatten() / tscal

    q0 = getvector(q0)
    qf = getvector(qf)

    if not len(q0) == len(qf):
        raise ValueError('q0 and q1 must be same size')

    if qd0 is None:
        qd0 = np.zeros(q0.shape)
    else:
        qd0 = getvector(qd0)
        if not len(qd0) == len(q0):
            raise ValueError('qd0 has wrong size')
    if qd1 is None:
        qd1 = np.zeros(q0.shape)
    else:
        qd0 = getvector(qd0)
        if not len(qd1) == len(q0):
            raise ValueError('qd1 has wrong size')

    # compute the polynomial coefficients
    A = 6 * (qf - q0) - 3 * (qd1 + qd0) * tscal
    B = -15 * (qf - q0) + (8 * qd0 + 7 * qd1) * tscal
    C = 10 * (qf - q0) - (6 * qd0 + 4 * qd1) * tscal
    E = qd0 * tscal  # as the t vector has been normalized
    F = q0

    # n = len(q0)

    tt = np.array([t**5, t**4, t**3, t**2, t, np.ones(t.shape)]).T
    coeffs = np.array([A, B, C, np.zeros(A.shape), E, F])

    qt = tt @ coeffs

    # compute  velocity
    coeffs = np.array(
        [np.zeros(A.shape), 5 * A, 4 * B, 3 * C, np.zeros(A.shape), E])
    qdt = tt @ coeffs / tscal

    # compute  acceleration
    coeffs = np.array([
        np.zeros(A.shape), np.zeros(A.shape),
        20 * A, 12 * B, 6 * C, np.zeros(A.shape)])
    qddt = tt @ coeffs / tscal ** 2

    return namedtuple('jtraj', 't q qd qdd')(tt, qt, qdt, qddt)


def t1plot(tg, block=True):
    """
    Plot 1D trajectories

    :param tg: the trajectory to plot
    :type tg: namedtuple

    Plot the position, velocity and acceleration contained in the namedtuple
    ``tg``.

    :seealso: :func:`~tpoly`, :func:`~lspb`
    """

    plotargs = {'markersize': 3}
    textargs = {'fontsize': 12}

    plt.figure()
    ax = plt.subplot(3, 1, 1)

    # plot position
    if type(tg).__name__ == 'tpoly':
        ax.plot(tg.x, tg.y, '-o', **plotargs)

    elif type(tg).__name__ == 'lspb':
        # accel phase
        tf = tg.x[-1]
        k = tg.x <= tg.xblend
        ax.plot(tg.x[k], tg.y[k], 'ro-', **plotargs)

        # coast phase
        k = (tg.x > tg.xblend) & (tg.x <= (tf-tg.xblend))
        ax.plot(tg.x[k], tg.y[k], 'bo-', **plotargs)
        k = np.where(k)[0][0]
        ax.plot(tg.x[k-1:k+1], tg.y[k-1:k+1], 'b-', **plotargs)

        # decel phase
        k = tg.x > (tf - tg.xblend)
        ax.plot(tg.x[k], tg.y[k], 'go-', **plotargs)
        k = np.where(k)[0][0]
        ax.plot(tg.x[k-1:k+1], tg.y[k-1:k+1], 'g-', **plotargs)

        ax.grid(True)
    else:
        raise TypeError('unknown 1D trajectory tuple')

    ax.grid(True)

    if tg.istime:
        ax.set_ylabel('$s(t)$', **textargs)
    else:
        ax.set_ylabel('$s(k)$', **textargs)

    # plot velocity
    ax = plt.subplot(3, 1, 2)
    ax.plot(tg.x, tg.yd, '-o', **plotargs)
    ax.grid(True)

    if tg.istime:
        ax.set_ylabel('$ds/dt$', **textargs)
    else:
        ax.set_ylabel('$ds/dk$', **textargs)

    # plot acceleration
    ax = plt.subplot(3, 1, 3)
    ax.plot(tg.x, tg.ydd, '-o', **plotargs)
    ax.grid(True)
    if tg.istime:
        ax.set_ylabel('$ds^2/dt^2$', **textargs)
        ax.set_xlabel('t (seconds)')
    else:
        ax.set_ylabel('$ds^2/dk^2$', **textargs)
        ax.set_xlabel('k (step))')

    plt.show(block=block)


def qplot(q, t=None, block=True):
    """
    Plot robot joint angles

    :param q: joint angle trajectory
    :type q: numpy ndarray, shape=(M,N)
    :param t: time vector, optional
    :type t: numpy ndarray, shape=(M,)

    This is a convenience function to plot joint angle trajectories (MxN) for
    an N-axis robot, where each row represents one time step.

    - ``qplot(q)`` plots the joint angles versus row number.  If N==6 a
      conventional 6-axis robot is assumed, and the first three joints are
      shown as solid lines, the last three joints (wrist) are shown as dashed
      lines. A legend is also displayed.

    - ``qplot(q, t)`` as above but displays the joint angle trajectory versus
      time given the time vector T (Mx1).

    :seealso: :func:`jtraj`
    """
    assertmatrix(q)

    if t is None:
        t = np.arange(0, q.shape[0])

    n = q.shape[1]
    fig, ax = plt.subplots()
    if n == 6:
        plt.plot(t, q[:, 0:3])
        plt.plot(t, q[:, 3:6], '--')
    else:
        plt.plot(t, q)

    ax.legend([f"q{i+1}" for i in range(n)])

    plt.grid(True)
    ax.set_xlabel('Time (s)')
    ax.set_ylabel('Joint coordinates (rad,m)')
    ax.set_xlim(t[0], t[-1])

    plt.show(block=block)


# -------------------------------------------------------------------------- #

def ctraj(T0, T1, s):
    """
    Cartesian trajectory between two poses

    :param T0: initial pose
    :type T0: SE3
    :param T1: final pose
    :type T1: SE3
    :return T0: smooth path from ``T0`` to ``T1``
    :rtype: SE3

    ``ctraj(T0, T1, n)`` is a Cartesian trajectory from SE3 pose ``T0`` to
    ``T1`` with ``n`` points that follow a trapezoidal velocity profile along
    the path. The Cartesian trajectory is an SE3 instance containing ``n``
    values.

    ``ctraj(T0, T1, s)`` as above but the elements of ``s`` specify the
    fractional distance  along the path, and these values are in the
    range [0 1]. The i'th point corresponds to a distance ``s[i]`` along
    the path.

    Examples::

        >>> tg = ctraj(SE3.Rand(), SE3.Rand(), 20)
        >>> len(tg)
        20

    Notes:

    - In the second case ``s`` could be generated by a scalar trajectory
      generator such as ``tpoly`` or ``lspb`` (default).
    - Orientation interpolation is performed using unit-quaternion
      interpolation.

    Reference:

    - Robotics, Vision & Control, Sec 3.1.5,
      Peter Corke, Springer 2011

    :seealso: :func:`~roboticstoolbox.trajectory.lspb`,
        :func:`~spatialmath.unitquaternion.interp`
    """

    if isinstance(s, int):
        s = lspb(0, 1, s).y
    elif isvector(s):
        s = getvector(s)
    else:
        raise TypeError('bad argument for time, must be int or vector')

    return T1.interp(s, start=T0)


def cmstraj():
    pass


# -------------------------------------------------------------------------- #

def mstraj(
        viapoints, dt, tacc, qdmax=None, tsegment=None,
        q0=None, qd0=None, qdf=None, verbose=False):
    """
    Multi-segment multi-axis trajectory

    :param viapoints: A set of viapoints, one per row
    :type viapoints: ndarray(m,n)
    :param dt: time step
    :type dt: float (seconds)
    :param tacc: acceleration time (seconds)
    :type tacc: float
    :param qdmax: maximum speed, defaults to None
    :type qdmax: array_like(n) or float, optional
    :param tsegment: maximum time of each motion segment (seconds), defaults
        to None
    :type tsegment: array_like, optional
    :param q0: initial coordinates, defaults to first row of viapoints
    :type q0: array_like(n), optional
    :param qd0: inital  velocity, defaults to zero
    :type qd0: array_like(n), optional
    :param qdf: final  velocity, defaults to zero
    :type qdf: array_like(n), optional
    :param verbose: print debug information, defaults to False
    :type verbose: bool, optional
    :return: trajectory plus extra info
    :rtype: namedtuple

    Computes a trajectory for N axes moving smoothly through a set of
    viapoints.
    The motion comprises M segments:

    - The initial coordinates are the first row of ``viapoints`` or ``q0`` if
      provided.
    - The final coordinates are the last row of ``viapoints``
    - Each segment is linear motion and polynomial blends connect the
      viapoints.
    - All joints arrive at each via point at the same time, ie. the motion is
      coordinated across axes

    The time of the segments can be specified in two different ways:

    #. In terms of segment time where ``tsegment`` is an array of segment times
       which is the number of via points minus one::

            ``traj = mstraj(viapoints, dt, tacc, tsegment=TS)``

    #. Governed by the speed of the slowest axis for the segment.  The axis
       speed is a scalar (all axes have the same speed) or an N-vector of speed
       per axis::

            traj = mstraj(viapoints, dt, tacc, qdmax=SPEED)

    The return value is a namedtuple (named ``mstraj``) with elements:

        - ``t``  the time coordinate as a numpy ndarray, shape=(K,)
        - ``q``  the axis values as a numpy ndarray, shape=(K,N)
        - ``arrive`` a list of arrival times for each segment
        - ``info`` a list of named tuples, one per segment that describe the
          slowest axis, segment time,  and time stamp
        - ``via`` the passed set of via points

    The  trajectory proper is (``traj.t``, ``traj.q``).  The trajectory is a
    matrix has one row per time step, and one column per axis.

     Notes:

     - Only one of ``qdmag`` or ``tsegment`` can be specified
     - If ``tacc`` is greater than zero then the path smoothly accelerates
       between segments using a polynomial blend.  This means that the the via
       point is not actually reached.
     - The path length K is a function of the number of via
       points and the time or velocity limits that apply.
     - Can be used to create joint space trajectories where each axis is a
       joint coordinate.
     - Can be used to create Cartesian trajectories where the "axes"
       correspond to translation and orientation in RPY or Euler angle form.
     - If ``qdmax`` is a scalar then all axes are assumed to have the same
       maximum speed.

    :seealso: `lspb`, `ctraj`, `mtraj`
    """

    if q0 is None:
        q0 = viapoints[0, :]
        viapoints = viapoints[1:, :]
    else:
        q0 = getvector(q0)
        if not viapoints.shape[1] == len(q0):
            raise ValueError('WP and Q0 must have same number of columns')

    ns, nj = viapoints.shape
    Tacc = tacc

    if qdmax is not None and tsegment is not None:
        raise ValueError('cannot specify both qdmax and tsegment')

    if qdmax is None:
        if tsegment is None:
            raise ValueError('tsegment must be given if qdmax is not')

        if not len(tsegment) == ns:
            raise ValueError(
                'Length of TSEG does not match number of viapoints')

    if tsegment is None:

        # This is unreachable, left just in case
        if qdmax is None:  # pragma nocover
            raise ValueError('qdmax must be given if tsegment is not')

        if isinstance(qdmax, (int, float)):
            # if qdmax is a scalar assume all axes have the same speed
            qdmax = np.tile(qdmax, (nj,))
        else:
            qdmax = getvector(qdmax)

            if not len(qdmax) == nj:
                raise ValueError(
                    'Length of QDMAX does not match number of axes')

    if isinstance(Tacc, (int, float)):
        Tacc = np.tile(Tacc, (ns,))
    else:
        if not len(Tacc) == ns:
            raise ValueError('Tacc is wrong size')
    if qd0 is None:
        qd0 = np.zeros((nj,))
    else:
        if not len(qd0) == len(q0):
            raise ValueError('qd0 is wrong size')
    if qdf is None:
        qdf = np.zeros((nj,))
    else:
        if not len(qdf) == len(q0):
            raise ValueError('qdf is wrong size')

    # set the initial conditions
    q_prev = q0
    qd_prev = qd0

    clock = 0     # keep track of time
    arrive = np.zeros((ns,))   # record planned time of arrival at via points
    tg = np.zeros((0, nj))
    infolist = []
    info = namedtuple('mstraj_info', 'slowest segtime clock')

    def mrange(start, stop, step):
        """
        mrange(start, stop, step) behaves like MATLAB start:step:stop
        and includes the final value unlike range() or np.arange()
        """
        ret = []
        v = start
        while v <= stop:
            ret.append(v)
            v += step
        return np.r_[ret]

    for seg in range(0, ns):
        q_next = viapoints[seg, :]    # current target

        if verbose:  # pragma nocover
            print(f"------- segment {seg}: {q_prev} --> {q_next}")

        # set the blend time, just half an interval for the first segment

        tacc = Tacc[seg]

        tacc = math.ceil(tacc / dt) * dt
        tacc2 = math.ceil(tacc / 2 / dt) * dt
        if seg == 0:
            taccx = tacc2
        else:
            taccx = tacc

        # estimate travel time
        #    could better estimate distance travelled during the blend

        dq = q_next - q_prev    # total distance to move this segment

        # probably should iterate over the next section to get qb right...
        # while 1
        #   qd_next = (qnextnext - qnext)
        #   tb = abs(qd_next - qd) ./ qddmax;
        #   qb = f(tb, max acceleration)
        #   dq = q_next - q_prev - qb
        #   tl = abs(dq) ./ qdmax;

        if qdmax is not None:
            # qdmax is specified, compute slowest axis

            # qb = taccx * qdmax / 2       # distance moved during blend
            tb = taccx

            # convert to time
            tl = abs(dq) / qdmax
            # tl = abs(dq - qb) / qdmax
            tl = np.ceil(tl / dt) * dt

            # find the total time and slowest axis
            tt = tb + tl
            slowest = np.argmax(tt)
            tseg = tt[slowest]

            # best if there is some linear motion component
            if tseg <= 2*tacc:
                tseg = 2 * tacc

        elif tsegment is not None:
            # segment time specified, use that
            tseg = tsegment[seg]
            slowest = math.nan

        infolist.append(info(slowest, tseg, clock))

        # log the planned arrival time
        arrive[seg] = clock + tseg
        if seg > 0:
            arrive[seg] += tacc2

        if verbose:   # pragma nocover
            print(
                f"seg {seg}, distance {dq}, "
                "slowest axis {slowest}, time required {tseg}")

        # create the trajectories for this segment

        # linear velocity from qprev to qnext
        qd = dq / tseg

        # add the blend polynomial
        qb = jtraj(
            q0, q_prev + tacc2 * qd, mrange(0, taccx, dt),
            qd0=qd_prev, qd1=qd).q
        if verbose:    # pragma nocover
            print(qb)
        tg = np.vstack([tg, qb[1:, :]])

        clock = clock + taccx     # update the clock

        # add the linear part, from tacc/2+dt to tseg-tacc/2
        for t in mrange(tacc2 + dt, tseg - tacc2, dt):
            s = t / tseg
            q0 = (1 - s) * q_prev + s * q_next       # linear step
            print(t, s, q0)
            tg = np.vstack([tg, q0])
            clock += dt

        q_prev = q_next    # next target becomes previous target
        qd_prev = qd

    # add the final blend
    qb = jtraj(q0, q_next, mrange(0, tacc2, dt), qd0=qd_prev, qd1=qdf).q
    tg = np.vstack([tg, qb[1:, :]])

    infolist.append(info(None, tseg, clock))

    return namedtuple(
        'mstraj', 't q arrive info via')(
            dt * np.arange(0, tg.shape[0]), tg, arrive, infolist, viapoints)



from roboticstoolbox.models.DH.Panda import Panda
from roboticstoolbox.models.DH.Puma560 import Puma560
from roboticstoolbox.models.DH.Stanford import Stanford
from roboticstoolbox.models.DH.Ball import Ball
from roboticstoolbox.models.DH.Hyper import Hyper
from roboticstoolbox.models.DH.Coil import Coil
from roboticstoolbox.models.DH.Cobra600 import Cobra600
from roboticstoolbox.models.DH.IRB140 import IRB140
from roboticstoolbox.models.DH.KR5 import KR5
from roboticstoolbox.models.DH.Orion5 import Orion5
from roboticstoolbox.models.DH.Planar3 import Planar3
from roboticstoolbox.models.DH.Planar2 import Planar2
from roboticstoolbox.models.DH.LWR4 import LWR4
from roboticstoolbox.models.DH.Sawyer import Sawyer
from roboticstoolbox.models.DH.UR3 import UR3
from roboticstoolbox.models.DH.UR5 import UR5
from roboticstoolbox.models.DH.UR10 import UR10
from roboticstoolbox.models.DH.Mico import Mico
from roboticstoolbox.models.DH.Jaco import Jaco


__all__ = [
    'Panda',
    'Puma560',
    'Stanford',
    'Ball',
    'Coil',
    'Hyper',
    'Cobra600',
    'IRB140',
    'KR5',
    'Orion5',
    'Planar3',
    'Planar2',
    'LWR4',
    'UR3',
    'UR5',
    'UR10',
    'Sawyer',
    'Mico',
    'Jaco',
]



#!/usr/bin/env python
"""
@author: Peter Corke
@author: Samuel Drew
"""

from roboticstoolbox import DHRobot, RevoluteDH, PrismaticDH
from math import pi
import numpy as np


class Cobra600(DHRobot):
    """
    Class that models a Adept Cobra 600 SCARA manipulator

    ``Cobra600()`` is a class which models an Adept Cobra 600 SCARA robot and
    describes its kinematic characteristics using standard DH
    conventions.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.DH.Cobra600()
        >>> print(robot)

    Defined joint configurations are:

    - qz, zero joint angle configuration, 'L' shaped configuration

    .. note::
        - SI units are used.
        - Robot has only 4 DoF.

    .. codeauthor:: Peter Corke
    """

    def __init__(self):
        deg = pi/180

        L = [RevoluteDH(d=0.387, a=0.325, qlim=[-50*deg, 50*deg]),
             RevoluteDH(a=0.275, alpha=pi, qlim=[-88*deg, 88*deg]),
             PrismaticDH(qlim=[0, 0.210]),
             RevoluteDH()]

        super().__init__(L, name='Cobra600', manufacturer='Adept')

        # zero angles, L shaped pose
        self.addconfiguration("qz", np.array([0, 0, 0, 0]))


if __name__ == '__main__':   # pragma nocover

    cobra = Cobra600()
    print(cobra)



#!/usr/bin/env python

import numpy as np
from roboticstoolbox.robot.ERobot import ERobot
from math import pi


class Puma560(ERobot):
    """
    Class that imports a Puma 560 URDF model

    ``Puma560()`` is a class which imports a Unimation Puma560 robot definition
    from a URDF file.  The model describes its kinematic and graphical
    characteristics.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.URDF.Puma560()
        >>> print(robot)

    Defined joint configurations are:

    - qz, zero joint angle configuration, 'L' shaped configuration
    - qr, vertical 'READY' configuration
    - qs, arm is stretched out in the x-direction
    - qn, arm is at a nominal non-singular configuration

    .. codeauthor:: Jesse Haviland
    .. sectionauthor:: Peter Corke
    """

    def __init__(self):

        args = super().urdf_to_ets_args(
            "puma560_description/urdf/puma560_robot.urdf.xacro")

        super().__init__(
            args[0],
            name=args[1])

        self.manufacturer = "Unimation"
        # self.ee_link = self.ets[9]

        # zero angles, L shaped pose
        self.addconfiguration("qz", np.array([0, 0, 0, 0, 0, 0]))

        # ready pose, arm up
        self.addconfiguration("qr", np.array([0, pi/2, -pi/2, 0, 0, 0]))

        # straight and horizontal
        self.addconfiguration("qs", np.array([0, 0, -pi/2, 0, 0, 0]))

        # nominal table top picking pose
        self.addconfiguration("qn", np.array([0, pi/4, pi, 0, pi/4, 0]))


if __name__ == '__main__':   # pragma nocover

    robot = Puma560()
    print(robot)



from roboticstoolbox.robot.Robot import Robot
from roboticstoolbox.robot.Link import Link
from roboticstoolbox.robot.DHRobot import SerialLink, DHRobot
from roboticstoolbox.robot.DHLink import DHLink, RevoluteDH, PrismaticDH, \
    RevoluteMDH, PrismaticMDH
from roboticstoolbox.robot.ERobot import ERobot
from roboticstoolbox.robot.ELink import ELink
from roboticstoolbox.robot.ETS import ETS, ETS2
from roboticstoolbox.robot.Gripper import Gripper
from roboticstoolbox.robot.Shape import Shape, Cylinder, Sphere, Box, Mesh

__all__ = [
    'Robot',
    'SerialLink',
    'DHRobot',
    'Link',
    'DHLink',
    'RevoluteDH',
    'PrismaticDH',
    'RevoluteMDH',
    'PrismaticMDH',
    'ERobot',
    'ELink',
    'ERobot',
    'ETS',
    'ETS2',
    'Gripper',
    'Shape',
    'Cylinder',
    'Sphere',
    'Box',
    'Mesh'
    ]



from roboticstoolbox.tools.urdf.urdf import (
    URDFType,
    Box, Cylinder, Sphere, Mesh, Geometry,
    Collision, Visual, Inertial,
    JointCalibration, JointDynamics, JointLimit, JointMimic,
    SafetyController, Actuator, TransmissionJoint,
    Transmission, Joint, Link, URDF)

__all__ = [
    'URDFType', 'Box', 'Cylinder', 'Sphere', 'Mesh', 'Geometry',
    'Collision', 'Visual', 'Inertial',
    'JointCalibration', 'JointDynamics', 'JointLimit', 'JointMimic',
    'SafetyController', 'Actuator', 'TransmissionJoint',
    'Transmission', 'Joint', 'Link', 'URDF'
]
